@@ -1,19 +1,29 @@
 /*
  * collector_sql.h
  *
- * Copyright (c) 2010-2011, NIPPON TELEGRAPH AND TELEPHONE CORPORATION
+ * Copyright (c) 2010-2012, NIPPON TELEGRAPH AND TELEPHONE CORPORATION
  */
 
 #ifndef COLLECTOR_SQL_H
 #define COLLECTOR_SQL_H
 
+#if PG_VERSION_NUM >= 90200
+#define PG_STAT_ACTIVITY_ATTNAME_PID		"pid"
+#define PG_STAT_ACTIVITY_ATTNAME_QUERY		"query"
+#define PG_STAT_REPLICATION_ATTNAME_PID		"pid"
+#else
+#define PG_STAT_ACTIVITY_ATTNAME_PID		"procpid"
+#define PG_STAT_ACTIVITY_ATTNAME_QUERY		"current_query"
+#define PG_STAT_REPLICATION_ATTNAME_PID		"procpid"
+#endif
+
 /*----------------------------------------------------------------------------
  * snapshots per instance
  *----------------------------------------------------------------------------
  */
 
 /* database */
-#if PG_VERSION_NUM >= 90100
+#if PG_VERSION_NUM >= 90200
 #define SQL_SELECT_DATABASE "\
 SELECT \
 	d.oid AS dbid, \
@@ -33,12 +43,48 @@ SELECT \
 	pg_stat_get_db_conflict_lock(d.oid) AS confl_lock, \
 	pg_stat_get_db_conflict_snapshot(d.oid) AS confl_snapshot, \
 	pg_stat_get_db_conflict_bufferpin(d.oid) AS confl_bufferpin, \
-	pg_stat_get_db_conflict_startup_deadlock(d.oid) AS confl_deadlock \
+	pg_stat_get_db_conflict_startup_deadlock(d.oid) AS confl_deadlock, \
+	pg_stat_get_db_temp_files(d.oid) AS temp_files, \
+	pg_stat_get_db_temp_bytes(d.oid) AS temp_bytes, \
+	pg_stat_get_db_deadlocks(d.oid) AS deadlocks, \
+	pg_stat_get_db_blk_read_time(d.oid) AS blk_read_time, \
+	pg_stat_get_db_blk_write_time(d.oid) AS blk_write_time \
 FROM \
 	pg_database d \
 WHERE datallowconn \
   AND datname <> ALL (('{' || $1 || '}')::text[]) \
-ORDER BY 1"
+ORDER BY dbid"
+#elif PG_VERSION_NUM >= 90100
+#define SQL_SELECT_DATABASE "\
+SELECT \
+	d.oid AS dbid, \
+	d.datname, \
+	pg_database_size(d.oid), \
+	CASE WHEN pg_is_in_recovery() THEN 0 ELSE age(d.datfrozenxid) END, \
+	pg_stat_get_db_xact_commit(d.oid) AS xact_commit, \
+	pg_stat_get_db_xact_rollback(d.oid) AS xact_rollback, \
+	pg_stat_get_db_blocks_fetched(d.oid) - pg_stat_get_db_blocks_hit(d.oid) AS blks_read, \
+	pg_stat_get_db_blocks_hit(d.oid) AS blks_hit, \
+	pg_stat_get_db_tuples_returned(d.oid) AS tup_returned, \
+	pg_stat_get_db_tuples_fetched(d.oid) AS tup_fetched, \
+	pg_stat_get_db_tuples_inserted(d.oid) AS tup_inserted, \
+	pg_stat_get_db_tuples_updated(d.oid) AS tup_updated, \
+	pg_stat_get_db_tuples_deleted(d.oid) AS tup_deleted, \
+	pg_stat_get_db_conflict_tablespace(d.oid) AS confl_tablespace, \
+	pg_stat_get_db_conflict_lock(d.oid) AS confl_lock, \
+	pg_stat_get_db_conflict_snapshot(d.oid) AS confl_snapshot, \
+	pg_stat_get_db_conflict_bufferpin(d.oid) AS confl_bufferpin, \
+	pg_stat_get_db_conflict_startup_deadlock(d.oid) AS confl_deadlock, \
+	NULL AS temp_files, \
+	NULL AS temp_bytes, \
+	NULL AS deadlocks, \
+	NULL AS blk_read_time, \
+	NULL AS blk_write_time \
+FROM \
+	pg_database d \
+WHERE datallowconn \
+  AND datname <> ALL (('{' || $1 || '}')::text[]) \
+ORDER BY dbid"
 #elif PG_VERSION_NUM >= 90000
 #define SQL_SELECT_DATABASE "\
 SELECT \
@@ -55,16 +101,21 @@ SELECT \
 	pg_stat_get_db_tuples_inserted(d.oid) AS tup_inserted, \
 	pg_stat_get_db_tuples_updated(d.oid) AS tup_updated, \
 	pg_stat_get_db_tuples_deleted(d.oid) AS tup_deleted, \
-	NULL::bigint AS confl_tablespace, \
-	NULL::bigint AS confl_lock, \
-	NULL::bigint AS confl_snapshot, \
-	NULL::bigint AS confl_bufferpin, \
-	NULL::bigint AS confl_deadlock \
+	NULL AS confl_tablespace, \
+	NULL AS confl_lock, \
+	NULL AS confl_snapshot, \
+	NULL AS confl_bufferpin, \
+	NULL AS confl_deadlock, \
+	NULL AS temp_files, \
+	NULL AS temp_bytes, \
+	NULL AS deadlocks, \
+	NULL AS blk_read_time, \
+	NULL AS blk_write_time \
 FROM \
 	pg_database d \
 WHERE datallowconn \
   AND datname <> ALL (('{' || $1 || '}')::text[]) \
-ORDER BY 1"
+ORDER BY dbid"
 #else
 #define SQL_SELECT_DATABASE "\
 SELECT \
@@ -81,16 +132,21 @@ SELECT \
 	pg_stat_get_db_tuples_inserted(d.oid) AS tup_inserted, \
 	pg_stat_get_db_tuples_updated(d.oid) AS tup_updated, \
 	pg_stat_get_db_tuples_deleted(d.oid) AS tup_deleted, \
-	NULL::bigint AS confl_tablespace, \
-	NULL::bigint AS confl_lock, \
-	NULL::bigint AS confl_snapshot, \
-	NULL::bigint AS confl_bufferpin, \
-	NULL::bigint AS confl_deadlock \
+	NULL AS confl_tablespace, \
+	NULL AS confl_lock, \
+	NULL AS confl_snapshot, \
+	NULL AS confl_bufferpin, \
+	NULL AS confl_deadlock, \
+	NULL AS temp_files, \
+	NULL AS temp_bytes, \
+	NULL AS deadlocks, \
+	NULL AS blk_read_time, \
+	NULL AS blk_write_time \
 FROM \
 	pg_database d \
 WHERE datallowconn \
   AND datname <> ALL (('{' || $1 || '}')::text[]) \
-ORDER BY 1"
+ORDER BY dbid"
 #endif
 
 /* activity */
@@ -133,50 +189,90 @@ FROM \
 	pg_roles"
 
 /* statement */
-#if PG_VERSION_NUM < 90000
-
+#if PG_VERSION_NUM >= 90200
 #define SQL_SELECT_STATEMENT "\
 SELECT \
-	dbid, \
-	userid, \
-	query, \
-	calls, \
-	total_time, \
-	rows, \
-	NULL::bigint AS shared_blks_hit, \
-	NULL::bigint AS shared_blks_read, \
-	NULL::bigint AS shared_blks_written, \
-	NULL::bigint AS local_blks_hit, \
-	NULL::bigint AS local_blks_read, \
-	NULL::bigint AS local_blks_written, \
-	NULL::bigint AS temp_blks_read, \
-	NULL::bigint AS temp_blks_written \
+	s.dbid, \
+	s.userid, \
+	s.query, \
+	s.calls, \
+	s.total_time / 1000, \
+	s.rows, \
+	s.shared_blks_hit, \
+	s.shared_blks_read, \
+	s.shared_blks_dirtied, \
+	s.shared_blks_written, \
+	s.local_blks_hit, \
+	s.local_blks_read, \
+	s.local_blks_dirtied, \
+	s.local_blks_written, \
+	s.temp_blks_read, \
+	s.temp_blks_written, \
+	s.blk_read_time, \
+	s.blk_write_time \
 FROM \
-	pg_stat_statements \
-ORDER BY total_time DESC LIMIT 30"
-
+	pg_stat_statements s \
+	LEFT JOIN pg_roles r ON r.oid = s.userid \
+WHERE \
+	r.rolname <> ALL (('{' || $1 || '}')::text[]) \
+ORDER BY \
+	s.total_time DESC LIMIT $2"
+#elif PG_VERSION_NUM >= 90100
+#define SQL_SELECT_STATEMENT "\
+SELECT \
+	s.dbid, \
+	s.userid, \
+	s.query, \
+	s.calls, \
+	s.total_time, \
+	s.rows, \
+	s.shared_blks_hit, \
+	s.shared_blks_read, \
+	NULL AS shared_blks_dirtied, \
+	s.shared_blks_written, \
+	s.local_blks_hit, \
+	s.local_blks_read, \
+	NULL AS local_blks_dirtied, \
+	s.local_blks_written, \
+	s.temp_blks_read, \
+	s.temp_blks_written, \
+	NULL AS blk_read_time, \
+	NULL AS blk_write_time \
+FROM \
+	pg_stat_statements s \
+	LEFT JOIN pg_roles r ON r.oid = s.userid \
+WHERE \
+	r.rolname <> ALL (('{' || $1 || '}')::text[]) \
+ORDER BY \
+	s.total_time DESC LIMIT $2"
 #else
-
 #define SQL_SELECT_STATEMENT "\
 SELECT \
-	dbid, \
-	userid, \
-	query, \
-	calls, \
-	total_time, \
-	rows, \
-	shared_blks_hit, \
-	shared_blks_read, \
-	shared_blks_written, \
-	local_blks_hit, \
-	local_blks_read, \
-	local_blks_written, \
-	temp_blks_read, \
-	temp_blks_written \
+	s.dbid, \
+	s.userid, \
+	s.query, \
+	s.calls, \
+	s.total_time, \
+	s.rows, \
+	NULL AS shared_blks_hit, \
+	NULL AS shared_blks_read, \
+	NULL AS shared_blks_dirtied, \
+	NULL AS shared_blks_written, \
+	NULL AS local_blks_hit, \
+	NULL AS local_blks_read, \
+	NULL AS local_blks_dirtied, \
+	NULL AS local_blks_written, \
+	NULL AS temp_blks_read, \
+	NULL AS temp_blks_written, \
+	NULL AS blk_read_time, \
+	NULL AS blk_write_time \
 FROM \
-	pg_stat_statements \
-ORDER BY total_time DESC LIMIT 30"
-
+	pg_stat_statements s \
+	LEFT JOIN pg_roles r ON r.oid = s.userid \
+WHERE \
+	r.rolname <> ALL (('{' || $1 || '}')::text[]) \
+ORDER BY \
+	s.total_time DESC LIMIT $2"
 #endif
 
 /* lock */
@@ -185,7 +281,7 @@ ORDER BY total_time DESC LIMIT 30"
 #define SQL_SELECT_LOCK_BLOCKER_QUERY "\
 CASE \
 	WHEN la.gid IS NOT NULL THEN '(xact is detached from session)' \
-	WHEN la.queries is null THEN '(library might not have been loaded)' \
+	WHEN la.queries IS NULL THEN '(library might not have been loaded)' \
 	ELSE la.queries \
 END"
 #else
@@ -216,7 +312,7 @@ SELECT \
 	la.pid AS blocker_pid, \
 	la.gid AS blocker_gid, \
 	(statement_timestamp() - sb.query_start)::interval(0), \
-	sb.current_query, \
+	sb." PG_STAT_ACTIVITY_ATTNAME_QUERY ", \
 	" SQL_SELECT_LOCK_BLOCKER_QUERY " \
 FROM \
 	(SELECT DISTINCT l0.pid, l0.relation, " SQL_SELECT_LOCK_XID_CAST ", la.gid, lx.queries \
@@ -229,11 +325,11 @@ FROM \
 		LEFT JOIN statsinfo.last_xact_activity() lx ON l0.pid = lx.pid \
 	 WHERE l0.granted = true AND \
 		(la.gid IS NULL OR l0.relation IS NOT NULL)) la \
-	 LEFT JOIN pg_stat_activity sa ON la.pid = sa.procpid, \
+	 LEFT JOIN pg_stat_activity sa ON la.pid = sa." PG_STAT_ACTIVITY_ATTNAME_PID ", \
 	(SELECT DISTINCT pid, relation, " SQL_SELECT_LOCK_XID_CAST " \
 	 FROM pg_locks \
 	 WHERE granted = false) lb \
-	 LEFT JOIN pg_stat_activity sb ON lb.pid = sb.procpid \
+	 LEFT JOIN pg_stat_activity sb ON lb.pid = sb." PG_STAT_ACTIVITY_ATTNAME_PID " \
 	 LEFT JOIN pg_database db ON sb.datid = db.oid \
 	 LEFT JOIN pg_class cb ON lb.relation = cb.oid \
 	 LEFT JOIN pg_namespace nb ON cb.relnamespace = nb.oid \
@@ -244,7 +340,7 @@ WHERE \
 /* replication */
 #define SQL_SELECT_REPLICATION "\
 SELECT \
-	procpid, \
+	" PG_STAT_REPLICATION_ATTNAME_PID ", \
 	usesysid, \
 	usename, \
 	application_name, \
@@ -263,6 +359,21 @@ SELECT \
 FROM \
 	pg_stat_replication"
 
+/* xlog */
+#if PG_VERSION_NUM >= 90000
+#define SQL_SELECT_XLOG "\
+SELECT \
+	pg_current_xlog_location(), \
+	pg_xlogfile_name(pg_current_xlog_location()) \
+WHERE \
+	NOT pg_is_in_recovery()"
+#else
+#define SQL_SELECT_XLOG "\
+SELECT \
+	pg_current_xlog_location(), \
+	pg_xlogfile_name(pg_current_xlog_location())"
+#endif
+
 /* cpu */
 #define SQL_SELECT_CPU "\
 SELECT * FROM statsinfo.cpustats($1)"
@@ -271,6 +382,14 @@ SELECT * FROM statsinfo.cpustats($1)"
 #define SQL_SELECT_DEVICE "\
 SELECT * FROM statsinfo.devicestats($1)"
 
+/* loadavg */
+#define SQL_SELECT_LOADAVG "\
+SELECT * FROM statsinfo.loadavg()"
+
+/* memory */
+#define SQL_SELECT_MEMORY "\
+SELECT * FROM statsinfo.memory()"
+
 /* profile */
 #define SQL_SELECT_PROFILE	"SELECT * FROM statsinfo.profile()"
 
@@ -386,7 +505,7 @@ SELECT \
 		WHEN s.stakind2 = 3 THEN s.stanumbers2[1] \
 		WHEN s.stakind3 = 3 THEN s.stanumbers3[1] \
 		WHEN s.stakind4 = 3 THEN s.stanumbers4[1] \
-		ELSE NULL::real \
+		ELSE NULL \
 	END AS correlation \
 FROM \
 	pg_attribute a \
