@@ -1,7 +1,7 @@
 /*
  * logger.c
  *
- * Copyright (c) 2010-2011, NIPPON TELEGRAPH AND TELEPHONE CORPORATION
+ * Copyright (c) 2010-2012, NIPPON TELEGRAPH AND TELEPHONE CORPORATION
  */
 
 #include "pg_statsinfod.h"
@@ -15,6 +15,10 @@
 #include <dirent.h>
 #endif
 
+/* pg_statsinfo.control values */
+static StatsinfoControlFileData		ControlFile;
+
+static int			cf_fd; 
 static time_t		logger_reload_time = 0;
 static List		   *adjust_log_list = NIL;
 
@@ -36,6 +40,7 @@ static char		   *my_adjust_log_warning;
 static char		   *my_adjust_log_error;
 static char		   *my_adjust_log_log;
 static char		   *my_adjust_log_fatal;
+static List		   *my_textlog_nologging_users;
 /*-----------------------*/
 
 #if PG_VERSION_NUM < 90000
@@ -87,20 +92,29 @@ static int				recursive_level = 0;
 static void reload_params(void);
 static void logger_recv(Logger *logger);
 
-static void logger_parse(Logger *logger, const char *pg_log);
+static void logger_parse(Logger *logger, const char *pg_log, bool only_routing);
 
 static void logger_route(Logger *logger, const Log *log);
 
 static void init_log(Log *log, const char *buf, size_t len, const size_t fields[]);
+static bool logger_open(Logger *logger, const char *csvlog);
 static void logger_close(Logger *logger);
 static bool logger_next(Logger *logger, const char *pg_log);
 static void get_csvlog(char csvlog[], const char *prev, const char *pg_log);
 static void adjust_log(Log *log);
 static void replace_extension(char path[], const char *extension);
 static void assign_textlog_path(Logger *logger, const char *pg_log);
-static void assign_csvlog_path(Logger *logger, const char *pg_log, const char *csvlog);
+static void assign_csvlog_path(Logger *logger, const char *pg_log, const char *csvlog, long offset);
 static List *add_adlog(List *adlog_list, int elevel, char *rawstring);
 static char *b_trim(char *str);
+static bool split_string(char *rawstring, char separator, List **elemlist);
+static bool is_nologging_user(const Log *log);
+static int csvfilter(const struct dirent *dp);
+static void open_controlfile(Logger *logger, int flags, mode_t mode);
+static void load_controlfile(Logger *logger);
+static bool ReadControlFile(void);
+static void RewriteControlFile(Logger *logger);
+static void logger_shutdown(Logger *logger);
 
 void
 logger_init(void)
@@ -112,19 +126,41 @@ logger_init(void)
 }
 
 /*
- * logger_main(const char *prev_csv_name)
+ * logger_main
  */
 void *
 logger_main(void *arg)
 {
-	Logger		logger;
-	const char *prev_csv_name = (const char *) arg;
+	Logger			  logger;
+	int				  entry;
+	struct dirent	**dplist;
 
 	memset(&logger, 0, sizeof(logger));
 	initStringInfo(&logger.buf);
 	reload_params();
+
+	/* sets latest csvlog to the elements of logger */
+	for (;;)
+	{
+		entry = scandir(my_log_directory, &dplist, csvfilter, alphasort);
+		if (entry > 0)
+			break;
+		usleep(200 * 1000);	/* 200ms */
+	};
+	assign_csvlog_path(&logger, my_log_directory, dplist[entry - 1]->d_name, 0);
 	assign_textlog_path(&logger, my_log_directory);
-	assign_csvlog_path(&logger, my_log_directory, prev_csv_name);
+
+	while (entry--)
+		free(dplist[entry]);
+	free(dplist);
+
+	/* load the pg_statsinfo.control */
+	load_controlfile(&logger);
+	ControlFile.state = STATSINFO_RUNNING;
+
+	/* perform the log routing until end of the latest csvlog */
+	logger_open(&logger, logger.csv_path);
+	logger_parse(&logger, my_log_directory, true);
 
 	/*
 	 * Logger should not shutdown before any other threads are alive,
@@ -150,7 +186,7 @@ logger_main(void *arg)
 				assign_textlog_path(&logger, my_log_directory);
 		}
 
-		logger_parse(&logger, my_log_directory);
+		logger_parse(&logger, my_log_directory, false);
 		usleep(200 * 1000);	/* 200ms */
 
 		/* check postmaster pid. */
@@ -167,7 +203,7 @@ logger_main(void *arg)
 
 		for (;;)
 		{
-			logger_parse(&logger, my_log_directory);
+			logger_parse(&logger, my_log_directory, false);
 			logger_recv(&logger);
 			if (shutdown_message_found || time(NULL) > until)
 				break;
@@ -182,10 +218,14 @@ logger_main(void *arg)
 		elog(WARNING, "shutdown because server process exited abnormally");
 	logger_recv(&logger);
 
+	/* update pg_statsinfo.control */
+	ControlFile.state = STATSINFO_SHUTDOWNED;
+	RewriteControlFile(&logger);
+
 	logger_close(&logger);
 	shutdown_progress(LOGGER_SHUTDOWN);
 
-	return NULL;
+	return (void *) LOGGER_RETURN_SUCCESS;
 }
 
 #ifdef PGUT_OVERRIDE_ELOG
@@ -374,6 +414,8 @@ reload_params(void)
 	 */
 	my_textlog_filename = pgut_strdup(textlog_filename);
 	my_textlog_permission = textlog_permission & 0666;
+	list_destroy(my_textlog_nologging_users, free);
+	split_string(textlog_nologging_users, ',', &my_textlog_nologging_users);
 
 	/* syslog */
 	my_syslog_min_messages = syslog_min_messages;
@@ -468,14 +510,18 @@ logger_route(Logger *logger, const Log *log)
 
 		if (logger->textlog != NULL)
 		{
-			if (!write_textlog(log,
-							   my_textlog_line_prefix,
-							   my_log_error_verbosity,
-							   logger->textlog))
+			/* don't write a textlog of the users that are set not logging */
+			if (!is_nologging_user(log))
 			{
-				/* unexpected error; close the file, and try to reopen */
-				fclose(logger->textlog);
-				logger->textlog = NULL;
+				if (!write_textlog(log,
+								   my_textlog_line_prefix,
+								   my_log_error_verbosity,
+								   logger->textlog))
+				{
+					/* unexpected error; close the file, and try to reopen */
+					fclose(logger->textlog);
+					logger->textlog = NULL;
+				}
 			}
 		}
 	}
@@ -521,7 +567,7 @@ init_log(Log *log, const char *buf, size_t len, const size_t fields[])
  * logger_parse - Parse CSV log and route it into textlog, syslog, or trap.
  */
 static void
-logger_parse(Logger *logger, const char *pg_log)
+logger_parse(Logger *logger, const char *pg_log, bool only_routing)
 {
 	while (logger_next(logger, pg_log))
 	{
@@ -534,38 +580,55 @@ logger_parse(Logger *logger, const char *pg_log)
 		if (log.elevel == LOG)
 		{
 			/* checkpoint ? */
-			if (parse_checkpoint(log.message, log.timestamp))
+			if (is_checkpoint(log.message))
+			{
+				if (!only_routing)
+					parse_checkpoint(log.message, log.timestamp);
 				continue;
+			}
 
 			/* autovacuum ? */
-			if (parse_autovacuum(log.message, log.timestamp))
+			if (is_autovacuum(log.message))
+			{
+				if (!only_routing)
+					parse_autovacuum(log.message, log.timestamp);
 				continue;
+			}
 
 			/* snapshot requested ? */
 			if (strcmp(log.message, LOGMSG_SNAPSHOT) == 0)
 			{
-				pthread_mutex_lock(&reload_lock);
-				free((char *) snapshot_requested);
-				snapshot_requested = pgut_strdup(log.detail);
-				pthread_mutex_unlock(&reload_lock);
+				if (!only_routing)
+				{
+					pthread_mutex_lock(&reload_lock);
+					free((char *) snapshot_requested);
+					snapshot_requested = pgut_strdup(log.detail);
+					pthread_mutex_unlock(&reload_lock);
+				}
 				continue;
 			}
 
 			/* maintenance requested ? */
 			if (strcmp(log.message, LOGMSG_MAINTENANCE) == 0)
 			{
-				pthread_mutex_lock(&reload_lock);
-				free((char *) maintenance_requested);
-				maintenance_requested = pgut_strdup(log.detail);
-				pthread_mutex_unlock(&reload_lock);
+				if (!only_routing)
+				{
+					pthread_mutex_lock(&reload_lock);
+					free((char *) maintenance_requested);
+					maintenance_requested = pgut_strdup(log.detail);
+					pthread_mutex_unlock(&reload_lock);
+				}
 				continue;
 			}
 
 			/* restart requested ? */
 			if (strcmp(log.message, LOGMSG_RESTART) == 0)
 			{
-				shutdown_message_found = true;
-				shutdown_progress(SHUTDOWN_REQUESTED);
+				if (!only_routing)
+				{
+					shutdown_message_found = true;
+					shutdown_progress(SHUTDOWN_REQUESTED);
+				}
 				continue;
 			}
 
@@ -590,7 +653,13 @@ logger_parse(Logger *logger, const char *pg_log)
 			adjust_log(&log);
 		logger_route(logger, &log);
 
-		if (save_elevel == LOG)
+		/* update pg_statsinfo.control */
+		strlcpy(ControlFile.csv_name,
+			logger->csv_name, sizeof(ControlFile.csv_name));
+		ControlFile.csv_offset = logger->csv_offset;
+		RewriteControlFile(logger);
+
+		if (!only_routing && save_elevel == LOG)
 		{
 			/* setting parameters reloaded ? */
 			if (strcmp(log.message, msg_sighup) == 0)
@@ -618,6 +687,35 @@ logger_parse(Logger *logger, const char *pg_log)
 	}
 }
 
+static bool
+logger_open(Logger *logger, const char *csvlog)
+{
+	mode_t	mask;
+
+	Assert(!logger->fp);
+	Assert(!logger->textlog);
+
+	assign_csvlog_path(logger, my_log_directory, csvlog, 0);
+	assign_textlog_path(logger, my_log_directory);
+
+	/* open csvlog file */
+	logger->fp = pgut_fopen(logger->csv_path, "rt");
+	if (logger->fp == NULL)
+		return false;
+
+	/* create a new textlog file */
+	mask = umask(0777 & ~my_textlog_permission);
+	logger->textlog = pgut_fopen(logger->textlog_path, "at");
+	umask(mask);
+	if (logger->textlog == NULL)
+	{
+		fclose(logger->fp);
+		return false;
+	}
+
+	return true;
+}
+
 static void
 logger_close(Logger *logger)
 {
@@ -637,9 +735,10 @@ logger_close(Logger *logger)
 		logger->textlog = NULL;
 
 		/* overwrite existing .log file; it must be empty.
-		 * (Note : Some error messages through system() (eg. recovery_command)
-		 *  outputs to the .log file  from postgres's logger, so sometimes
-		 *  .log file will not be empty. At the moment we overwrite without check.
+		 * Note:
+		 * Some error messages through system() (eg. recovery_command)
+		 * outputs to the .log file  from postgres's logger, so sometimes
+		 * .log file will not be empty. At the moment we overwrite without check.
 		 */
 		if (logger->csv_path[0] && stat(logger->csv_path, &st) == 0)
 		{
@@ -667,7 +766,6 @@ logger_next(Logger *logger, const char *pg_log)
 	{
 		char	csvlog[MAXPGPATH];
 		char	textlog[MAXPGPATH];
-		mode_t	mask;
 
 		if (shutdown_message_found)
 			return false;	/* must end with the current log */
@@ -685,8 +783,9 @@ logger_next(Logger *logger, const char *pg_log)
 		/*
 		 * csvlog files that have empty *.log have not been parsed yet
 		 * because postgres logger make an empty log file.
-		 * (Note: Some "cannot stat" error messages are output to *.log,
-		 *  so we check logger->fp and csvlog again.
+		 * Note:
+		 * Some "cannot stat" error messages are output to *.log,
+		 * so we check logger->fp and csvlog again.
 		 */
 		 if (stat(textlog, &st) == 0 && st.st_size > 0)
 		 {
@@ -695,18 +794,9 @@ logger_next(Logger *logger, const char *pg_log)
 		 }
 
 		logger_close(logger);
-		assign_textlog_path(logger, pg_log);
-		assign_csvlog_path(logger, pg_log, csvlog);
-
-		logger->fp = pgut_fopen(logger->csv_path, "rt");
-		if (logger->fp == NULL)
+		if (!logger_open(logger, csvlog))
 			return false;
 
-		/* create a new textlog file */
-		mask = umask(0777 & ~my_textlog_permission);
-		logger->textlog = pgut_fopen(logger->textlog_path, "at");
-		umask(mask);
-
 		elog(DEBUG2, "read csvlog \"%s\"", logger->csv_path);
 	}
 
@@ -741,8 +831,11 @@ logger_next(Logger *logger, const char *pg_log)
 static void
 get_csvlog(char csvlog[], const char *prev, const char *pg_log)
 {
-	DIR			   *dir;
-	struct dirent  *dp;
+	int				  entry;
+	struct dirent	**dplist;
+	struct stat		  st;
+	char			  tmppath[MAXPGPATH];
+	int				  i;
 
 	Assert(csvlog);
 	Assert(prev);
@@ -751,32 +844,32 @@ get_csvlog(char csvlog[], const char *prev, const char *pg_log)
 
 	csvlog[0] = '\0';
 
-	if ((dir = opendir(pg_log)) == NULL)
+	if ((entry = scandir(pg_log, &dplist, csvfilter, alphasort)) < 0)
 	{
 		/* pg_log directory might not exist before syslogger started */
 		if (errno != ENOENT)
 			ereport(WARNING,
 				(errcode_errno(),
-				 errmsg("could not open directory \"%s\": ", pg_log)));
+				 errmsg("could not scan directory \"%s\": ", pg_log)));
 		return;
 	}
 
-	for (dp = readdir(dir); dp != NULL; dp = readdir(dir))
+	for (i = 0; i < entry; i++)
 	{
-		const char *extension = strrchr(dp->d_name, '.');
-
-		/* check the extension is .csv */
-		if (extension == NULL || strcmp(extension, ".csv") != 0)
-			continue;
-
 		/* get the next log of previous parsed log */
-		if (strcmp(prev, dp->d_name) >= 0 ||
-			(csvlog[0] && strcmp(csvlog, dp->d_name) < 0))
-			continue;
+		if (strcmp(prev, dplist[i]->d_name) < 0)
+		{
+			strlcpy(csvlog, dplist[i]->d_name, MAXPGPATH);
 
-		strlcpy(csvlog, dp->d_name, MAXPGPATH);
+			join_path_components(tmppath, pg_log, dplist[i]->d_name);
+			if (stat(tmppath, &st) == 0 && st.st_size > 0)
+				break;
+		}
 	}
-	closedir(dir);
+
+	while (entry--)
+		free(dplist[entry]);
+	free(dplist);
 }
 
 static void
@@ -818,11 +911,15 @@ assign_textlog_path(Logger *logger, const char *pg_log)
 }
 
 static void
-assign_csvlog_path(Logger *logger, const char *pg_log, const char *csvlog)
+assign_csvlog_path(Logger *logger, const char *pg_log, const char *csvlog, long offset)
 {
-	join_path_components(logger->csv_path, pg_log, csvlog);
-	logger->csv_name = logger->csv_path + strlen(pg_log) + 1;
-	logger->csv_offset = 0;
+	if (is_absolute_path(csvlog))
+		strlcpy(logger->csv_path, csvlog, MAXPGPATH);
+	else
+		join_path_components(logger->csv_path, pg_log, csvlog);
+
+	logger->csv_name = last_dir_separator(logger->csv_path) + 1;
+	logger->csv_offset = offset;
 }
 
 static List *
@@ -866,3 +963,324 @@ b_trim(char *str)
 
 	return str;
 }
+
+/*
+ * Note: this function modify the argument
+ */
+static bool
+split_string(char *rawstring, char separator, List **elemlist)
+{
+	char	*nextp = rawstring;
+	bool	 done = false;
+
+	*elemlist = NIL;
+
+	/* skip leading whitespace */
+	while (isspace((unsigned char) *nextp))
+		nextp++;
+
+	/* allow empty string */
+	if (*nextp == '\0')
+		return true;
+
+	/* At the top of the loop, we are at start of a new identifier. */
+	do
+	{
+		char *curname;
+		char *endp;
+
+		if (*nextp == '\"')
+		{
+			/* Quoted name --- collapse quote-quote pairs, no downcasing */
+			curname = nextp + 1;
+			for (;;)
+			{
+				endp = strchr(nextp + 1, '\"');
+				if (endp == NULL)
+					return false; /* mismatched quotes */
+				if (endp[1] != '\"')
+					break; /* found end of quoted name */
+				/* Collapse adjacent quotes into one quote, and look again */
+				memmove(endp, endp + 1, strlen(endp));
+				nextp = endp;
+			}
+			/* endp now points at the terminating quote */
+			nextp = endp + 1;
+		}
+		else
+		{
+			/* Unquoted name --- extends to separator or whitespace */
+			curname = nextp;
+			while (*nextp && *nextp != separator &&
+				   !isspace((unsigned char) *nextp))
+				nextp++;
+			endp = nextp;
+			if (curname == nextp)
+				return false; /* empty unquoted name not allowed */
+		}
+
+		/* skip trailing whitespace */
+		while (isspace((unsigned char) *nextp))
+			nextp++;
+
+		if (*nextp == separator)
+		{
+			nextp++;
+			while (isspace((unsigned char) *nextp))
+				nextp++; /* skip leading whitespace for next */
+			/* we expect another name, so done remains false */
+		}
+		else if (*nextp == '\0')
+			done = true;
+		else
+			return false; /* invalid syntax */
+
+		/* Now safe to overwrite separator with a null */
+		*endp = '\0';
+
+		/*
+		 * Finished isolating current name --- add it to list
+		 */
+		*elemlist = lappend(*elemlist, pgut_strdup(curname));
+
+		/* Loop back if we didn't reach end of string */
+	} while (!done);
+
+	return true;
+}
+
+static bool
+is_nologging_user(const Log *log)
+{
+	ListCell	*cell;
+
+	foreach(cell, my_textlog_nologging_users)
+	{
+		if (strcmp(log->username, (char *) lfirst(cell)) == 0)
+		return true;
+	}
+	return false;
+}
+
+static int
+csvfilter(const struct dirent *dp)
+{
+	const char	*extension;
+
+	if (dp->d_type != DT_REG)
+		return 0;
+
+	/* check the extension is .csv */
+	extension = strrchr(dp->d_name, '.');
+	if (extension && strcmp(extension, ".csv") == 0)
+		return 1;
+	else
+		return 0;
+}
+
+static void
+open_controlfile(Logger *logger, int flags, mode_t mode)
+{
+	if ((cf_fd = open(STATSINFO_CONTROL_FILE, flags, mode)) < 0)
+	{
+		ereport(ERROR,
+			(errcode_errno(),
+			 errmsg("could not open control file \"%s\": %m",
+			 	STATSINFO_CONTROL_FILE)));
+		/* shutdown due to fatal error */
+		logger_shutdown(logger);
+	}
+}
+
+/*
+ * load the previous state from pg_statsinfo.control.
+ */
+static void
+load_controlfile(Logger *logger)
+{
+	struct stat		st;
+
+	if (stat(STATSINFO_CONTROL_FILE, &st) != 0)
+	{
+		open_controlfile(logger,
+			O_RDWR | O_CREAT | O_EXCL | PG_BINARY,
+			S_IRUSR | S_IWUSR);
+		return;
+	}
+
+	open_controlfile(logger, O_RDWR | PG_BINARY, 0);
+
+	/*
+	 * if state value of pg_statsinfo.control is not "STATSINFO_SHUTDOWNED",
+	 * that means the previous pg_statsinfod was abnormally end.
+	 * in that case, parse the csvlog between csvlog of the point of abnormal
+	 * termination and latest csvlog.
+	 */
+	if (ReadControlFile())
+	{
+		char prev_csvlog[MAXPGPATH];
+
+		if (ControlFile.state == STATSINFO_SHUTDOWNED)
+			return;
+
+		join_path_components(
+			prev_csvlog, my_log_directory, ControlFile.csv_name);
+
+		if (stat(prev_csvlog, &st) == 0 && ControlFile.csv_offset <= st.st_size)
+		{
+			/* set the csvlog path and the csvlog offset to the logger */
+			assign_csvlog_path(logger, my_log_directory,
+			ControlFile.csv_name, ControlFile.csv_offset);
+			return;
+		}
+
+		/* csvlog which parsed at last is missed */
+		ereport(WARNING,
+			(errmsg("csvlog file \"%s\" not found or incurrect offset",
+				ControlFile.csv_name)));
+	}
+
+	/*
+	 * could not read the pg_statsinfo.control or incorrect data.
+	 * rename the latest textlog file to "<latest-csvlog>.err.<seqid>"
+	 * (eg. postgresql-2012-07-01_000000.err.1)
+	 */
+	if (stat(logger->textlog_path, &st) == 0)
+	{
+		char new_path[MAXPGPATH];
+		char extension[32];
+		int seqid = 0;
+
+		for (;;)
+		{
+			strlcpy(new_path, logger->csv_path, sizeof(new_path));
+			snprintf(extension, sizeof(extension), ".err.%d", ++seqid);
+			replace_extension(new_path, extension);
+
+			if (stat(new_path, &st) != 0)
+				break;
+		}
+
+		rename(logger->textlog_path, new_path);
+		elog(WARNING,
+			"latest textlog file already exists, it renamed to '%s'", new_path);
+	}
+}
+
+/*
+ * I/O routines for pg_statsinfo.control
+ *
+ * ControlFile is a buffer in memory that holds an image of the contents of
+ * pg_statsinfo.control. RewriteControlFile() writes the pg_statsinfo.control
+ * file with the contents in buffer. ReadControlFile() loads the buffer from the
+ * pg_statsinfo.control file.
+ */
+static bool
+ReadControlFile(void)
+{
+	pg_crc32	crc;
+
+	Assert(cf_fd >= 0);	/* have not been opened the pg_statsinfo.control */
+
+	/* read data */
+	if (read(cf_fd, &ControlFile,
+			sizeof(ControlFile)) != sizeof(ControlFile))
+	{
+		ereport(ERROR,
+			(errcode_errno(),
+			 errmsg("could not read from control file \"%s\": %m",
+			 	STATSINFO_CONTROL_FILE)));
+		return false;
+	}
+
+	/*
+	 * Check for expected pg_statsinfo.control format version.
+	 * If this is wrong, the CRC check will likely fail because we'll be
+	 * checking the wrong number of bytes.
+	 * Complaining about wrong version will probably be more enlightening
+	 * than complaining about wrong CRC.
+	 */
+	if (ControlFile.control_version != STATSINFO_CONTROL_VERSION &&
+		((ControlFile.control_version / 100) != (STATSINFO_CONTROL_VERSION / 100)))
+	{
+		ereport(ERROR,
+			(errmsg("pg_statsinfo.control format incompatible"),
+			 errdetail("pg_statsinfo.control was created with STATSINFO_CONTROL_VERSION %d (0x%08x), "
+			 		   "but the pg_statsinfo was compiled with STATSINFO_CONTROL_VERSION %d (0x%08x)",
+					ControlFile.control_version, ControlFile.control_version,
+					STATSINFO_CONTROL_VERSION, STATSINFO_CONTROL_VERSION)));
+		return false;
+	}
+
+	/* check the CRC */
+	INIT_CRC32(crc);
+	COMP_CRC32(crc,
+		(char *) &ControlFile, offsetof(StatsinfoControlFileData, crc));
+	FIN_CRC32(crc);
+
+	if (!EQ_CRC32(crc, ControlFile.crc))
+	{
+		ereport(ERROR,
+			(errmsg("incorrect checksum in control file \"%s\"",
+				STATSINFO_CONTROL_FILE)));
+		return false;
+	}
+
+	return true;
+}
+
+static void
+RewriteControlFile(Logger *logger)
+{
+	char	buffer[sizeof(ControlFile)];
+
+	Assert(cf_fd >= 0);	/* have not been opened the pg_statsinfo.control */
+
+	/* initialize version and compatibility-check fields */
+	ControlFile.control_version = STATSINFO_CONTROL_VERSION;
+
+	/* contents are protected with a CRC */
+	INIT_CRC32(ControlFile.crc);
+	COMP_CRC32(ControlFile.crc,
+		(char *) &ControlFile, offsetof(StatsinfoControlFileData, crc));
+	FIN_CRC32(ControlFile.crc);
+
+	memset(buffer, 0, sizeof(ControlFile));
+	memcpy(buffer, &ControlFile, sizeof(ControlFile));
+
+	lseek(cf_fd, 0, SEEK_SET);
+	errno = 0;
+	if (write(cf_fd, buffer, sizeof(ControlFile)) != sizeof(ControlFile))
+	{
+		/* if write didn't set errno, assume problem is no disk space */
+		if (errno == 0)
+			errno = ENOSPC;
+		ereport(ERROR,
+			(errcode_errno(),
+			 errmsg("could not write to control file \"%s\": %m",
+			 	STATSINFO_CONTROL_FILE)));
+		/* shutdown due to fatal error */
+		logger_shutdown(logger);
+	}
+}
+
+static void
+logger_shutdown(Logger *logger)
+{
+	/* notify shutdown request to other threads */
+	if (shutdown_state < SHUTDOWN_REQUESTED)
+		shutdown_progress(SHUTDOWN_REQUESTED);
+
+	/* wait until the end of the other threads */
+	while (shutdown_state < WRITER_SHUTDOWN)
+		usleep(200 * 1000);	/* 200ms */
+
+	/* final shutdown message */
+	elog(ERROR, "shutdown due to fatal error");
+	logger_recv(logger);
+	logger_close(logger);
+
+	/* exit the logger thread */
+	shutdown_progress(LOGGER_SHUTDOWN);
+	pthread_exit((void *) LOGGER_RETURN_FAILED);
+}
