@@ -1,7 +1,7 @@
 /*
  * pg_statsinfod.c
  *
- * Copyright (c) 2010-2011, NIPPON TELEGRAPH AND TELEPHONE CORPORATION
+ * Copyright (c) 2010-2012, NIPPON TELEGRAPH AND TELEPHONE CORPORATION
  */
 
 #include "pg_statsinfod.h"
@@ -10,7 +10,7 @@
 
 #include "miscadmin.h"
 
-const char *PROGRAM_VERSION	= "2.3.1";
+const char *PROGRAM_VERSION	= "2.4.0";
 const char *PROGRAM_URL		= "http://pgstatsinfo.projects.postgresql.org/";
 const char *PROGRAM_EMAIL	= "pgstatsinfo-general@pgfoundry.org";
 
@@ -26,7 +26,6 @@ char		   *instance_id;			/* system identifier */
 static pid_t	postmaster_pid;			/* postmaster's pid */
 char		   *postmaster_port;		/* postmaster port number as string */
 static char	   *share_path;				/* $PGHOME/share */
-static char	   *prev_csv_name;			/* virtual previous csv name */
 int				server_version_num;		/* PG_VERSION_NUM */
 char		   *server_version_string;	/* PG_VERSION */
 int				server_encoding = -1;	/* server character encoding */
@@ -35,6 +34,8 @@ char		   *log_timezone_name;
 char		   *data_directory;
 char		   *excluded_dbnames;
 char		   *excluded_schemas;
+char		   *stat_statements_max;
+char		   *stat_statements_exclude_users;
 int				sampling_interval;
 int				snapshot_interval;
 /*---- GUC variables (logger) ----------*/
@@ -55,11 +56,13 @@ char		   *adjust_log_warning;
 char		   *adjust_log_error;
 char		   *adjust_log_log;
 char		   *adjust_log_fatal;
+char		   *textlog_nologging_users;
 /*---- GUC variables (writer) ----------*/
 char		   *repository_server;
-bool		    enable_maintenance;
+int			    enable_maintenance;
 time_t			maintenance_time;
 int				repository_keepday;
+char		   *log_maintenance_command;
 /*---- message format ----*/
 char		   *msg_debug;
 char		   *msg_info;
@@ -99,10 +102,13 @@ static bool assign_syslog(const char *value, void *var);
 static bool assign_string(const char *value, void *var);
 static bool assign_bool(const char *value, void *var);
 static bool assign_time(const char *value, void *var);
+static bool connect_test(const char *conninfo);
+static void readopt(void);
 static void after_readopt(void);
 static bool decode_time(const char *field, int *hour, int *min, int *sec);
 static int strtoi(const char *nptr, char **endptr, int base);
 static bool execute_script(PGconn *conn, const char *script_file);
+static bool check_repository(void);
 
 /* parameters */
 static struct ParamMap PARAM_MAP[] =
@@ -111,7 +117,6 @@ static struct ParamMap PARAM_MAP[] =
 	{"postmaster_pid", assign_int, &postmaster_pid},
 	{"port", assign_string, &postmaster_port},
 	{"share_path", assign_string, &share_path},
-	{"prev_csv_name", assign_string, &prev_csv_name},
 	{"server_version_num", assign_int, &server_version_num},
 	{"server_version_string", assign_string, &server_version_string},
 	{"server_encoding", assign_int, &server_encoding},
@@ -123,6 +128,8 @@ static struct ParamMap PARAM_MAP[] =
 	{"syslog_ident", assign_string, &syslog_ident},
 	{GUC_PREFIX ".excluded_dbnames", assign_string, &excluded_dbnames},
 	{GUC_PREFIX ".excluded_schemas", assign_string, &excluded_schemas},
+	{GUC_PREFIX ".stat_statements_max", assign_string, &stat_statements_max},
+	{GUC_PREFIX ".stat_statements_exclude_users", assign_string, &stat_statements_exclude_users},
 	{GUC_PREFIX ".repository_server", assign_string, &repository_server},
 	{GUC_PREFIX ".sampling_interval", assign_int, &sampling_interval},
 	{GUC_PREFIX ".snapshot_interval", assign_int, &snapshot_interval},
@@ -139,9 +146,11 @@ static struct ParamMap PARAM_MAP[] =
 	{GUC_PREFIX ".adjust_log_error", assign_string, &adjust_log_error},
 	{GUC_PREFIX ".adjust_log_log", assign_string, &adjust_log_log},
 	{GUC_PREFIX ".adjust_log_fatal", assign_string, &adjust_log_fatal},
-	{GUC_PREFIX ".enable_maintenance", assign_bool, &enable_maintenance},
+	{GUC_PREFIX ".textlog_nologging_users", assign_string, &textlog_nologging_users},
+	{GUC_PREFIX ".enable_maintenance", assign_int, &enable_maintenance},
 	{GUC_PREFIX ".maintenance_time", assign_time, &maintenance_time},
 	{GUC_PREFIX ".repository_keepday", assign_int, &repository_keepday},
+	{GUC_PREFIX ".log_maintenance_command", assign_string, &log_maintenance_command},
 	{":debug", assign_string, &msg_debug},
 	{":info", assign_string, &msg_info},
 	{":notice", assign_string, &msg_notice},
@@ -175,6 +184,8 @@ isTTY(int fd)
 int
 main(int argc, char *argv[])
 {
+	void	*logger_retval;
+
 	shutdown_state = STARTUP;
 
 	pgut_init(argc, argv);
@@ -183,9 +194,9 @@ main(int argc, char *argv[])
 	if (isTTY(fileno(stdin)))
 		return help();
 
-	/* read required parameters from stdin */
-	readopt_from_file(stdin);
-	fclose(stdin);
+	/* read required parameters */
+	readopt();
+
 	if (instance_id == NULL ||
 		postmaster_pid == 0 ||
 		postmaster_port == NULL ||
@@ -193,7 +204,6 @@ main(int argc, char *argv[])
 		data_directory == NULL ||
 		log_directory == NULL ||
 		share_path == NULL ||
-		prev_csv_name == NULL ||
 		msg_shutdown == NULL ||
 		msg_shutdown_smart == NULL ||
 		msg_shutdown_fast == NULL ||
@@ -238,6 +248,10 @@ main(int argc, char *argv[])
 	 */
 	pgut_abort_level = FATAL;
 
+	/* check the state of repository database */
+	if (!check_repository())
+		return STATSINFO_EXIT_FAILED;
+
 	/* init logger, collector, and writer module */
 	pthread_mutex_init(&shutdown_state_lock, NULL);
 	collector_init();
@@ -249,19 +263,22 @@ main(int argc, char *argv[])
 	elog(LOG, "start");
 	pthread_create(&th_collector, NULL, collector_main, NULL);
 	pthread_create(&th_writer, NULL, writer_main, NULL);
-	pthread_create(&th_logger, NULL, logger_main, prev_csv_name);
+	pthread_create(&th_logger, NULL, logger_main, NULL);
 
 	/* join the threads */ 
-	pthread_join(th_collector, NULL);
-	pthread_join(th_writer, NULL);
-	pthread_join(th_logger, NULL);
+	pthread_join(th_collector, (void **) NULL);
+	pthread_join(th_writer, (void **) NULL);
+	pthread_join(th_logger, (void **) &logger_retval);
 
 #ifdef NOT_USED
 	if (!shutdown_message_found)
 		restart_postmaster();		/* postmaster might be crashed! */
 #endif
 
-	return 0;
+	if (logger_retval == (void *) LOGGER_RETURN_FAILED)
+		return STATSINFO_EXIT_FAILED;
+
+	return STATSINFO_EXIT_SUCCESS;
 }
 
 static int
@@ -283,7 +300,7 @@ postmaster_is_alive(void)
 {
 #ifdef WIN32
 	static HANDLE hProcess = NULL;
-	
+
 	if (hProcess == NULL)
 	{
 		hProcess = OpenProcess(SYNCHRONIZE, false, postmaster_pid);
@@ -573,6 +590,59 @@ assign_param(const char *name, const char *value)
 	return true;
 }
 
+static bool
+connect_test(const char *conninfo)
+{
+#if PG_VERSION_NUM >= 90100
+	return PQping(conninfo) == PQPING_OK;
+#else
+	PGconn			*conn;
+	ConnStatusType	 status;
+
+	conn = PQconnectdb(conninfo);
+	status = PQstatus(conn);
+	PQfinish(conn);
+	return status == CONNECTION_OK;
+#endif
+}
+
+/*
+ * read required parameters.
+ */
+static void
+readopt(void)
+{
+	PGconn		*conn;
+	PGresult	*res;
+	char		 conninfo[1024];
+
+	/* read required parameters from stdin */
+	readopt_from_file(stdin);
+	fclose(stdin);
+
+	Assert(postmaster_port);
+
+	snprintf(conninfo, lengthof(conninfo),
+		"dbname=%s port=%s options='-c log_statement=none'",
+		"postgres", postmaster_port);
+
+	/* wait until postmaster is ready to accept connection */
+	while (postmaster_is_alive())
+	{
+		if (connect_test(conninfo))
+			break;
+		sleep(1);	/* 1s */
+	}
+
+	/* read required parameters from db */
+	conn = pgut_connect(conninfo, NO, ERROR);
+	res = pgut_execute(conn, SQL_SELECT_CUSTOM_SETTINGS, 0, NULL);
+	readopt_from_db(res);
+
+	PQclear(res);
+	pgut_disconnect(conn);
+}
+
 /*
  * Assign options from file. The file format must be:
  *	uint32	name_size
@@ -651,6 +721,10 @@ readopt_from_db(PGresult *res)
 	}
 
 	after_readopt();
+
+	if (!(enable_maintenance & MAINTENANCE_MODE_SNAPSHOT))
+		elog(NOTICE,
+			"automatic maintenance is disable. Please note the data size of the repository");
 }
 
 /*
@@ -670,10 +744,6 @@ after_readopt(void)
 		if (flags)
 			checkpoint_starting_prefix_len = flags - msg_checkpoint_starting;
 	}
-
-	if (!enable_maintenance)
-		elog(NOTICE,
-			"automatic maintenance is disable. Please note the data size of the repository");
 }
 
 /*
@@ -850,3 +920,60 @@ execute_script(PGconn *conn, const char *script_file)
 	termStringInfo(&buf);
 	return ok;
 }
+
+/*
+ * check the state of repository database.
+ *  - connect to the repository database
+ *  - statsrepo schema version
+ */
+static bool
+check_repository(void)
+{
+	PGconn		*conn;
+	PGresult	*res;
+	uint32		 version;
+
+	/* connect to the repository database */
+	if ((conn = pgut_connect(repository_server, NO, ERROR)) == NULL)
+		return false;
+
+	/* check statsrepo schema exists */
+	res = pgut_execute(conn,
+		"SELECT nspname FROM pg_namespace WHERE nspname = 'statsrepo'", 0, NULL);
+	if (PQresultStatus(res) != PGRES_TUPLES_OK)
+		goto error;
+	if (PQntuples(res) == 0)
+		goto done;	/* not installed */
+	PQclear(res);
+
+	/* check the statsrepo schema version */
+	res = pgut_execute(conn, "SELECT statsrepo.get_version()", 0, NULL);
+	if (PQresultStatus(res) != PGRES_TUPLES_OK)
+	{
+		char *sqlstate;
+
+		sqlstate = PQresultErrorField(res, PG_DIAG_SQLSTATE);
+		if (sqlstate && strcmp(sqlstate, "42883") == 0)	/* undefined function */
+			elog(ERROR, "incompatible statsrepo schema: version mismatch");
+		goto error;
+	}
+
+	/* verify the version of statsrepo schema */
+	parse_uint32(PQgetvalue(res, 0, 0), &version);
+	if (version != STATSREPO_SCHEMA_VERSION &&
+		((version / 100) != (STATSREPO_SCHEMA_VERSION / 100)))
+	{
+		elog(ERROR, "incompatible statsrepo schema: version mismatch");
+		goto error;
+	}
+
+done:
+	PQclear(res);
+	pgut_disconnect(conn);
+	return true;
+
+error:
+	PQclear(res);
+	pgut_disconnect(conn);
+	return false;
+}
