@@ -3,7 +3,7 @@
  *
  * Create a repository schema for PostgreSQL 8.3.
  *
- * Copyright (c) 2010-2011, NIPPON TELEGRAPH AND TELEPHONE CORPORATION
+ * Copyright (c) 2010-2012, NIPPON TELEGRAPH AND TELEPHONE CORPORATION
  */
 
 -- Adjust this setting to control where the objects get created.
@@ -73,6 +73,11 @@ CREATE TABLE statsrepo.database
 	confl_snapshot		bigint,
 	confl_bufferpin		bigint,
 	confl_deadlock		bigint,
+	temp_files			bigint,
+	temp_bytes			bigint,
+	deadlocks			bigint,
+	blk_read_time		double precision,
+	blk_write_time		double precision,
 	PRIMARY KEY (snapid, dbid),
 	FOREIGN KEY (snapid) REFERENCES statsrepo.snapshot (snapid) ON DELETE CASCADE
 );
@@ -242,12 +247,16 @@ CREATE TABLE statsrepo.statement
 	rows				bigint,
 	shared_blks_hit		bigint,
 	shared_blks_read	bigint,
+	shared_blks_dirtied	bigint,
 	shared_blks_written	bigint,
 	local_blks_hit		bigint,
 	local_blks_read		bigint,
+	local_blks_dirtied	bigint,
 	local_blks_written	bigint,
 	temp_blks_read		bigint,
 	temp_blks_written	bigint,
+	blk_read_time		double precision,
+	blk_write_time		double precision,
 	FOREIGN KEY (snapid) REFERENCES statsrepo.snapshot (snapid) ON DELETE CASCADE,
 	FOREIGN KEY (snapid, dbid) REFERENCES statsrepo.database (snapid, dbid)
 );
@@ -262,8 +271,8 @@ CREATE TABLE statsrepo.function
 	funcname		name,
 	argtypes		text,
 	calls			bigint,
-	total_time		bigint,
-	self_time		bigint,
+	total_time		double precision,
+	self_time		double precision,
 	FOREIGN KEY (snapid) REFERENCES statsrepo.snapshot (snapid) ON DELETE CASCADE,
 	FOREIGN KEY (snapid, dbid) REFERENCES statsrepo.database (snapid, dbid)
 );
@@ -281,6 +290,11 @@ CREATE TABLE statsrepo.autovacuum
 	page_remain		integer,
 	tup_removed		bigint,
 	tup_remain		bigint,
+	page_hit		integer,
+	page_miss		integer,
+	page_dirty		integer,
+	read_rate		double precision,
+	write_rate		double precision,
 	duration		real,
 	FOREIGN KEY (instid) REFERENCES statsrepo.instance (instid) ON DELETE CASCADE
 );
@@ -352,6 +366,28 @@ CREATE TABLE statsrepo.device
 	FOREIGN KEY (snapid) REFERENCES statsrepo.snapshot (snapid) ON DELETE CASCADE
 );
 
+CREATE TABLE statsrepo.loadavg
+(
+	snapid			bigint,
+	loadavg1		real,
+	loadavg5		real,
+	loadavg15		real,
+	PRIMARY KEY (snapid),
+	FOREIGN KEY (snapid) REFERENCES statsrepo.snapshot (snapid) ON DELETE CASCADE
+);
+
+CREATE TABLE statsrepo.memory
+(
+	snapid		bigint,
+	memfree		bigint,
+	buffers		bigint,
+	cached		bigint,
+	swap		bigint,
+	dirty		bigint,
+	PRIMARY KEY (snapid),
+	FOREIGN KEY (snapid) REFERENCES statsrepo.snapshot (snapid) ON DELETE CASCADE
+);
+
 CREATE TABLE statsrepo.profile
 (
 	snapid			bigint,
@@ -403,6 +439,14 @@ CREATE TABLE statsrepo.replication
 	FOREIGN KEY (snapid) REFERENCES statsrepo.snapshot (snapid) ON DELETE CASCADE
 );
 
+CREATE TABLE statsrepo.xlog
+(
+	snapid		bigint,
+	location	text,
+	xlogfile	text,
+	FOREIGN KEY (snapid) REFERENCES statsrepo.snapshot (snapid) ON DELETE CASCADE
+);
+
 -- del_snapshot(snapid) - delete the specified snapshot.
 CREATE FUNCTION statsrepo.del_snapshot(bigint) RETURNS void AS
 $$
@@ -433,6 +477,11 @@ $$ LANGUAGE sql;
 -- utility function for reporter.
 ------------------------------------------------------------------------------
 
+-- get_version() - version of statsrepo schema
+CREATE FUNCTION statsrepo.get_version() RETURNS text AS
+'SELECT CAST(''20400'' AS TEXT)'
+LANGUAGE sql IMMUTABLE;
+
 -- tps() - transaction per seconds
 CREATE FUNCTION statsrepo.tps(numeric, interval) RETURNS numeric AS
 'SELECT (CASE WHEN extract(epoch FROM $2) > 0 THEN $1 / extract(epoch FROM $2) ELSE 0 END)::numeric(1000, 3)'
@@ -448,6 +497,76 @@ CREATE FUNCTION statsrepo.sub(anyelement, anyelement) RETURNS anyelement AS
 'SELECT coalesce($1, 0) - coalesce($2, 0)'
 LANGUAGE sql;
 
+-- convert_hex() - convert a hexadecimal string to a decimal number
+CREATE FUNCTION statsrepo.convert_hex(text)
+RETURNS bigint AS
+$$
+	SELECT
+		(sum((16::numeric ^ (length($1) - i)) *
+			position(upper(substring($1 from i for 1)) in '123456789ABCDEF')))::bigint
+	FROM
+		generate_series(1, length($1)) AS t(i);
+$$
+LANGUAGE sql IMMUTABLE STRICT;
+
+-- xlog_location_diff() - compute the difference in bytes between two WAL locations
+CREATE FUNCTION statsrepo.xlog_location_diff(text, text)
+RETURNS numeric AS
+$$
+	/* XLogFileSize * (xlogid1 - xlogid2) + xrecoff1 - xrecoff2 */
+	SELECT
+		(X'FF000000'::bigint * (t.xlogid1 - t.xlogid2)::numeric + t.xrecoff1 - t.xrecoff2)
+	FROM
+	(
+		SELECT
+			statsrepo.convert_hex(lsn1[1]) AS xlogid1,
+			statsrepo.convert_hex(lsn1[2]) AS xrecoff1,
+			statsrepo.convert_hex(lsn2[1]) AS xlogid2,
+			statsrepo.convert_hex(lsn2[2]) AS xrecoff2
+		 FROM
+			regexp_matches($1, '^([0-F]{1,8})/([0-F]{1,8})$') AS lsn1,
+			regexp_matches($2, '^([0-F]{1,8})/([0-F]{1,8})$') AS lsn2
+	) t;
+$$
+LANGUAGE sql IMMUTABLE STRICT;
+
+-- pg_size_pretty() - formatting with size units
+CREATE FUNCTION statsrepo.pg_size_pretty(bigint)
+RETURNS text AS
+$$
+DECLARE
+	size	bigint := $1;
+	buf		text;
+	limit1	bigint := 10 * 1024;
+	limit2	bigint := limit1 * 2 - 1;
+BEGIN
+	IF size < limit1 THEN
+		buf := size || ' bytes';
+	ELSE
+		size := size >> 9;	/* keep one extra bit for rounding */
+		IF size < limit2 THEN
+			buf := (size + 1) / 2 || ' KiB';
+		ELSE
+			size := size >> 10;
+			IF size < limit2 THEN
+				buf := (size + 1) / 2 || ' MiB';
+			ELSE
+				size := size >> 10;
+				IF size < limit2 THEN
+					buf := (size + 1) / 2 || ' GiB';
+				ELSE
+					size := size >> 10;
+					buf := (size + 1) / 2 || ' TiB';
+				END IF;
+			END IF;
+		END IF;
+	END IF;
+
+	RETURN buf;
+END;
+$$
+LANGUAGE plpgsql IMMUTABLE STRICT;
+
 -- tables - pre-JOINed tables
 CREATE VIEW statsrepo.tables AS
   SELECT t.snapid,
@@ -832,7 +951,7 @@ $$
 		statsrepo.snapshot b,
 		statsrepo.snapshot e,
 		(SELECT
-			pg_size_pretty(sum(ed.size)::int8),
+			statsrepo.pg_size_pretty(sum(ed.size)::int8),
 			sum(ed.xact_commit) - sum(sd.xact_commit),
 			sum(ed.xact_rollback) - sum(sd.xact_rollback)
 		 FROM
@@ -861,7 +980,12 @@ CREATE FUNCTION statsrepo.get_dbstats(
 	OUT blks_hit_rate		numeric,
 	OUT blks_hit_tps		numeric,
 	OUT blks_read_tps		numeric,
-	OUT tup_fetch_tps		numeric
+	OUT tup_fetch_tps		numeric,
+	OUT temp_files			bigint,
+	OUT temp_bytes			bigint,
+	OUT deadlocks			bigint,
+	OUT blk_read_time		numeric,
+	OUT blk_write_time		numeric
 ) RETURNS SETOF record AS
 $$
 	SELECT
@@ -888,7 +1012,12 @@ $$
 		statsrepo.tps(
 			statsrepo.sub(ed.tup_returned, sd.tup_returned) +
 			statsrepo.sub(ed.tup_fetched, sd.tup_fetched),
-			es.time - ss.time)
+			es.time - ss.time),
+		statsrepo.sub(ed.temp_files, sd.temp_files),
+		statsrepo.sub(ed.temp_bytes, sd.temp_bytes) / 1024 / 1024,
+		statsrepo.sub(ed.deadlocks, sd.deadlocks),
+		statsrepo.sub(ed.blk_read_time, sd.blk_read_time)::numeric(1000, 3),
+		statsrepo.sub(ed.blk_write_time, sd.blk_write_time)::numeric(1000, 3)
 	FROM
 		statsrepo.snapshot ss,
 		statsrepo.snapshot es,
@@ -1178,10 +1307,14 @@ CREATE FUNCTION statsrepo.get_proc_tendency(
 $$
 	SELECT
 		a.snapid,
-		idle::numeric(1000, 3), 
-		idle_in_xact::numeric(1000, 3),
-		waiting::numeric(1000, 3),
-		running::numeric(1000, 3)
+		CASE WHEN (idle + idle_in_xact + waiting + running) = 0 THEN 0
+			ELSE (100 * idle / (idle + idle_in_xact + waiting + running))::numeric(5,2) END,
+		CASE WHEN (idle + idle_in_xact + waiting + running) = 0 THEN 0
+			ELSE (100 * idle_in_xact / (idle + idle_in_xact + waiting + running))::numeric(5,2) END,
+		CASE WHEN (idle + idle_in_xact + waiting + running) = 0 THEN 0
+			ELSE (100 * waiting / (idle + idle_in_xact + waiting + running))::numeric(5,2) END,
+		CASE WHEN (idle + idle_in_xact + waiting + running) = 0 THEN 0
+			ELSE (100 * running / (idle + idle_in_xact + waiting + running))::numeric(5,2) END
 	FROM
 		statsrepo.activity a,
 		statsrepo.snapshot s
@@ -1208,10 +1341,14 @@ CREATE FUNCTION statsrepo.get_proc_tendency_report(
 $$
 	SELECT
 		to_char(s.time, 'YYYY-MM-DD HH24:MI'),
-		idle::numeric(1000, 3), 
-		idle_in_xact::numeric(1000, 3),
-		waiting::numeric(1000, 3),
-		running::numeric(1000, 3)
+		CASE WHEN (idle + idle_in_xact + waiting + running) = 0 THEN 0
+			ELSE (100 * idle / (idle + idle_in_xact + waiting + running))::numeric(5,2) END,
+		CASE WHEN (idle + idle_in_xact + waiting + running) = 0 THEN 0
+			ELSE (100 * idle_in_xact / (idle + idle_in_xact + waiting + running))::numeric(5,2) END,
+		CASE WHEN (idle + idle_in_xact + waiting + running) = 0 THEN 0
+			ELSE (100 * waiting / (idle + idle_in_xact + waiting + running))::numeric(5,2) END,
+		CASE WHEN (idle + idle_in_xact + waiting + running) = 0 THEN 0
+			ELSE (100 * running / (idle + idle_in_xact + waiting + running))::numeric(5,2) END
 	FROM
 		statsrepo.activity a,
 		statsrepo.snapshot s
@@ -1225,6 +1362,84 @@ $$
 $$
 LANGUAGE sql;
 
+-- generate information that corresponds to 'WAL Statistics'
+CREATE FUNCTION statsrepo.get_xlog_tendency(
+	IN snapid_begin			bigint,
+	IN snapid_end			bigint,
+	OUT "timestamp"			text,
+	OUT location			text,
+	OUT xlogfile			text,
+	OUT write_size			numeric,
+	OUT write_size_per_sec	numeric
+) RETURNS SETOF record AS
+$$
+	SELECT
+		to_char(time, 'YYYY-MM-DD HH24:MI'),
+		location,
+		xlogfile,
+		(write_size / 1024 / 1024)::numeric(1000, 3),
+		(statsrepo.tps(write_size, duration) / 1024 / 1024)::numeric(1000, 3)
+	FROM
+	(
+		SELECT
+			xe.snapid,
+			xe.time,
+			xe.location,
+			xe.xlogfile,
+			statsrepo.xlog_location_diff(xe.location, xs.location) AS write_size,
+			xe.time - xs.time AS duration
+		FROM
+		 	(SELECT
+		 		s.snapid,
+		 		s.time,
+		 		x.location,
+		 		x.xlogfile,
+				(SELECT max(snapid) FROM statsrepo.snapshot WHERE snapid < s.snapid AND instid = s.instid) AS prev_snapid
+		 	 FROM
+			 	statsrepo.xlog x,
+				statsrepo.snapshot s
+			 WHERE
+				x.snapid BETWEEN $1 AND $2
+				AND x.snapid = s.snapid
+				AND s.instid = (SELECT instid FROM statsrepo.snapshot WHERE snapid = $2)) AS xe,
+		 	(SELECT
+		 		s.snapid,
+		 		s.time,
+		 		x.location,
+		 		x.xlogfile,
+				(SELECT min(snapid) FROM statsrepo.snapshot WHERE snapid > s.snapid AND instid = s.instid) AS next_snapid
+		 	 FROM
+			 	statsrepo.xlog x,
+				statsrepo.snapshot s
+			 WHERE
+				x.snapid BETWEEN $1 AND $2
+				AND x.snapid = s.snapid
+				AND s.instid = (SELECT instid FROM statsrepo.snapshot WHERE snapid = $2)) AS xs
+		WHERE
+			xs.snapid = xe.prev_snapid
+			AND xe.snapid = xs.next_snapid
+	) t
+	WHERE
+		snapid > $1;
+$$
+LANGUAGE sql;
+
+-- generate information that corresponds to 'WAL Statistics'
+CREATE FUNCTION statsrepo.get_xlog_stats(
+	IN snapid_begin		bigint,
+	IN snapid_end		bigint,
+	OUT write_total		numeric,
+	OUT write_speed		numeric
+) RETURNS SETOF record AS
+$$
+	SELECT
+		sum(write_size)::numeric(1000, 3),
+		avg(write_size_per_sec)::numeric(1000, 3)
+	FROM
+		statsrepo.get_xlog_tendency($1, $2);
+$$
+LANGUAGE sql;
+
 -- generate information that corresponds to 'CPU Usage'
 CREATE FUNCTION statsrepo.get_cpu_usage(
 	IN snapid_begin	bigint,
@@ -1683,6 +1898,64 @@ $$
 $$
 LANGUAGE sql;
 
+-- generate information that corresponds to 'Load Average'
+CREATE FUNCTION statsrepo.get_loadavg_tendency(
+	IN snapid_begin		bigint,
+	IN snapid_end		bigint,
+	OUT "timestamp"		text,
+	OUT "1min"			numeric,
+	OUT "5min"			numeric,
+	OUT "15min"			numeric
+) RETURNS SETOF record AS
+$$
+	SELECT
+		to_char(s.time, 'YYYY-MM-DD HH24:MI'),
+		l.loadavg1::numeric(1000, 3),
+		l.loadavg5::numeric(1000, 3),
+		l.loadavg15::numeric(1000, 3)
+	FROM
+		statsrepo.loadavg l,
+		statsrepo.snapshot s
+	WHERE
+		s.snapid BETWEEN $1 AND $2
+		AND s.snapid = l.snapid
+		AND s.instid = (SELECT instid FROM statsrepo.snapshot WHERE snapid = $2)
+	ORDER BY
+		s.snapid;
+$$
+LANGUAGE sql;
+
+-- generate information that corresponds to 'Memory Usage'
+CREATE FUNCTION statsrepo.get_memory_tendency(
+	IN snapid_begin		bigint,
+	IN snapid_end		bigint,
+	OUT "timestamp"		text,
+	OUT memfree			numeric,
+	OUT buffers			numeric,
+	OUT cached			numeric,
+	OUT swap			numeric,
+	OUT dirty			numeric
+) RETURNS SETOF record AS
+$$
+	SELECT
+		to_char(s.time, 'YYYY-MM-DD HH24:MI'),
+		(m.memfree::float / 1024)::numeric(1000, 2),
+		(m.buffers::float / 1024)::numeric(1000, 2),
+		(m.cached::float / 1024)::numeric(1000, 2),
+		(m.swap::float / 1024)::numeric(1000, 2),
+		(m.dirty::float / 1024)::numeric(1000, 2)
+	FROM
+		statsrepo.memory m,
+		statsrepo.snapshot s
+	WHERE
+		s.snapid BETWEEN $1 AND $2
+		AND s.snapid = m.snapid
+		AND s.instid = (SELECT instid FROM statsrepo.snapshot WHERE snapid = $2)
+	ORDER BY
+		s.snapid;
+$$
+LANGUAGE sql;
+
 -- generate information that corresponds to 'Disk Usage per Tablespace'
 CREATE FUNCTION statsrepo.get_disk_usage_tablespace(
 	IN snapid_begin		bigint,
@@ -1955,7 +2228,6 @@ $$
 		AND i.snapid = c.snapid
 		AND i.tbl = c.tbl
 		AND c.attnum = ANY (i.indkey)
-		AND i.isclustered
 		AND i.schema NOT IN ('pg_catalog', 'pg_toast', 'information_schema')
 		AND c.correlation < 1
 	ORDER BY
@@ -2033,6 +2305,43 @@ $$
 $$
 LANGUAGE sql;
 
+-- generate information that corresponds to 'Autovacuum Activity'
+CREATE FUNCTION statsrepo.get_autovacuum_activity2(
+	IN snapid_begin		bigint,
+	IN snapid_end		bigint,
+	OUT datname			text,
+	OUT nspname			text,
+	OUT relname			text,
+	OUT avg_page_hit	numeric,
+	OUT avg_page_miss	numeric,
+	OUT avg_page_dirty	numeric,
+	OUT avg_read_rate	numeric,
+	OUT avg_write_rate	numeric
+) RETURNS SETOF record AS
+$$
+	SELECT
+		database,
+		schema,
+		"table",
+		round(avg(page_hit)::numeric,3),
+		round(avg(page_miss)::numeric,3),
+		round(avg(page_dirty)::numeric,3),
+		round(avg(read_rate)::numeric,3),
+		round(avg(write_rate)::numeric,3)
+	FROM
+		statsrepo.autovacuum v,
+		(SELECT min(time) AS time FROM statsrepo.snapshot WHERE snapid >= $1) b,
+		(SELECT max(time) AS time FROM statsrepo.snapshot WHERE snapid <= $2) e
+	WHERE
+		v.start BETWEEN b.time AND e.time
+		AND v.instid = (SELECT instid FROM statsrepo.snapshot WHERE snapid = $2)
+	GROUP BY
+		database, schema, "table"
+	ORDER BY
+		4 DESC;
+$$
+LANGUAGE sql;
+
 -- generate information that corresponds to 'Query Activity (Functions)'
 CREATE FUNCTION statsrepo.get_query_activity_functions(
 	IN snapid_begin		bigint,
@@ -2042,8 +2351,8 @@ CREATE FUNCTION statsrepo.get_query_activity_functions(
 	OUT nspname			name,
 	OUT proname			name,
 	OUT calls			bigint,
-	OUT total_time		bigint,
-	OUT self_time		bigint,
+	OUT total_time		numeric,
+	OUT self_time		numeric,
 	OUT time_per_call	numeric
 ) RETURNS SETOF record AS
 $$
@@ -2053,10 +2362,10 @@ $$
 		s.name,
 		fe.funcname,
 		statsrepo.sub(fe.calls, fb.calls),
-		statsrepo.sub(fe.total_time, fb.total_time),
-		statsrepo.sub(fe.self_time, fb.self_time),
+		statsrepo.sub(fe.total_time, fb.total_time)::numeric(1000, 3),
+		statsrepo.sub(fe.self_time, fb.self_time)::numeric(1000, 3),
 		statsrepo.div(
-			statsrepo.sub(fe.total_time, fb.total_time),
+			statsrepo.sub(fe.total_time, fb.total_time)::numeric,
 			statsrepo.sub(fe.calls, fb.calls))
 	FROM
 		statsrepo.function fe LEFT JOIN statsrepo.function fb
@@ -2086,7 +2395,9 @@ CREATE FUNCTION statsrepo.get_query_activity_statements(
 	OUT query			text,
 	OUT calls			bigint,
 	OUT total_time		numeric,
-	OUT time_per_call	numeric
+	OUT time_per_call	numeric,
+	OUT blk_read_time	numeric,
+	OUT blk_write_time	numeric
 ) RETURNS SETOF record AS
 $$
 	SELECT
@@ -2097,7 +2408,9 @@ $$
 		statsrepo.sub(se.total_time, sb.total_time)::numeric(1000, 3),
 		statsrepo.div(
 			statsrepo.sub(se.total_time, sb.total_time)::numeric,
-			statsrepo.sub(se.calls, sb.calls))
+			statsrepo.sub(se.calls, sb.calls)),
+		statsrepo.sub(se.blk_read_time, sb.blk_read_time)::numeric(1000, 3),
+		statsrepo.sub(se.blk_write_time, sb.blk_write_time)::numeric(1000, 3)
 	FROM
 		statsrepo.statement se LEFT JOIN statsrepo.statement sb
 			ON sb.snapid = $1 AND sb.dbid = se.dbid AND sb.query = se.query,
@@ -2115,7 +2428,7 @@ $$
 $$
 LANGUAGE sql;
 
--- generate information that corresponds to 'Lock Activity'
+-- generate information that corresponds to 'Lock Conflicts'
 CREATE FUNCTION statsrepo.get_lock_activity(
 	IN snapid_begin		bigint,
 	IN snapid_end		bigint,
@@ -2161,7 +2474,7 @@ CREATE FUNCTION statsrepo.get_replication_activity(
 	OUT client_addr			inet,
 	OUT client_hostname		text,
 	OUT client_port			integer,
-	OUT backend_start		timestamptz,
+	OUT backend_start		timestamp,
 	OUT state				text,
 	OUT current_location	text,
 	OUT sent_location		text,
@@ -2178,7 +2491,7 @@ $$
 		client_addr,
 		client_hostname,
 		client_port,
-		backend_start,
+		backend_start::timestamp(0),
 		state,
 		current_location,
 		sent_location,
