@@ -1,7 +1,7 @@
 /*
  * lib/libstatsinfo.c
  *
- * Copyright (c) 2010-2011, NIPPON TELEGRAPH AND TELEPHONE CORPORATION
+ * Copyright (c) 2010-2012, NIPPON TELEGRAPH AND TELEPHONE CORPORATION
  */
 
 #include "libstatsinfo.h"
@@ -9,6 +9,8 @@
 #include <fcntl.h>
 #include <unistd.h>
 #include <sys/stat.h>
+#include <sys/types.h>
+#include <sys/wait.h>
 #include <time.h>
 
 #include "access/heapam.h"
@@ -17,23 +19,33 @@
 #include "catalog/pg_tablespace.h"
 #include "funcapi.h"
 #include "libpq/ip.h"
+#include "libpq/pqsignal.h"
 #include "mb/pg_wchar.h"
 #include "regex/regex.h"
 #include "miscadmin.h"
 #include "pgstat.h"
 #include "postmaster/autovacuum.h"
 #include "postmaster/syslogger.h"
+#include "postmaster/fork_process.h"
+#include "postmaster/postmaster.h"
+#include "storage/ipc.h"
 #include "storage/proc.h"
 #include "storage/procarray.h"
 #include "utils/builtins.h"
 #include "utils/guc.h"
 #include "utils/tqual.h"
 #include "utils/lsyscache.h"
+#include "utils/ps_status.h"
 
 #if PG_VERSION_NUM >= 90100
 #include "catalog/pg_collation.h"
 #endif
 
+#if PG_VERSION_NUM >= 90200
+#include "utils/timestamp.h"
+#include "utils/rel.h"
+#endif
+
 #include "pgut/pgut-be.h"
 #include "pgut/pgut-spi.h"
 #include "../common.h"
@@ -71,11 +83,21 @@
 	"received SIGHUP, reloading configuration files"
 
 /* log_autovacuum_min_duration: vacuum */
+#if PG_VERSION_NUM >= 90200
+#define MSG_AUTOVACUUM \
+	"automatic vacuum of table \"%s.%s.%s\": index scans: %d\n" \
+	"pages: %d removed, %d remain\n" \
+	"tuples: %.0f removed, %.0f remain\n" \
+	"buffer usage: %d hits, %d misses, %d dirtied\n" \
+	"avg read rate: %.3f MiB/s, avg write rate: %.3f MiB/s\n" \
+	"system usage: %s"
+#else
 #define MSG_AUTOVACUUM \
 	"automatic vacuum of table \"%s.%s.%s\": index scans: %d\n" \
 	"pages: %d removed, %d remain\n" \
 	"tuples: %.0f removed, %.0f remain\n" \
 	"system usage: %s"
+#endif
 
 /* log_autovacuum_min_duration: analyze */
 #define MSG_AUTOANALYZE \
@@ -101,18 +123,40 @@
 
 PG_MODULE_MAGIC;
 
+static const char *
+default_log_maintenance_command(void)
+{
+	char	bin_path[MAXPGPATH];
+	char	command[MAXPGPATH];
+
+	/* $PGHOME/bin */
+	strlcpy(bin_path, my_exec_path, MAXPGPATH);
+	get_parent_directory(bin_path);
+
+	snprintf(command, sizeof(command),
+		"%s/%s %%l", bin_path, "archive_pglog.sh");
+	return pstrdup(command);
+}
+
 /*---- GUC variables ----*/
 
 #define DEFAULT_SAMPLING_INTERVAL		5		/* sec */
 #define DEFAULT_SNAPSHOT_INTERVAL		600		/* sec */
 #define DEFAULT_SYSLOG_LEVEL			DISABLE
 #define DEFAULT_TEXTLOG_LEVEL			WARNING
-#define DEFAULT_ENABLE_MAINTENANCE		true
 #define DEFAULT_MAINTENANCE_TIME		"00:02:00"
 #define DEFAULT_REPOSITORY_KEEPDAY		7		/* day */
+#define DEFAULT_LOG_MAINTENANCE_COMMAND	default_log_maintenance_command()
 #define DEFAULT_LONG_LOCK_THREASHOLD	30		/* sec */
+#define DEFAULT_STAT_STATEMENTS_MAX		30
 #define LONG_TRANSACTION_THRESHOLD		1.0		/* sec */
 
+#if PG_VERSION_NUM < 80400
+#define DEFAULT_ENABLE_MAINTENANCE		"3"		/* snapshot + log */
+#else
+#define DEFAULT_ENABLE_MAINTENANCE		"on"	/* snapshot + log */
+#endif
+
 static const struct config_enum_entry elevel_options[] =
 {
 	{ "debug"	, DEBUG2 },
@@ -165,10 +209,14 @@ static char	   *adjust_log_warning = NULL;
 static char	   *adjust_log_error = NULL;
 static char	   *adjust_log_log = NULL;
 static char	   *adjust_log_fatal = NULL;
-static bool		enable_maintenance = DEFAULT_ENABLE_MAINTENANCE;
+static char	   *textlog_nologging_users = NULL;
+static char	   *enable_maintenance = NULL;
 static char	   *maintenance_time = NULL;
 static int		repository_keepday = DEFAULT_REPOSITORY_KEEPDAY;
+static char	   *log_maintenance_command = NULL;
 static int		long_lock_threashold = DEFAULT_LONG_LOCK_THREASHOLD;
+static int		stat_statements_max = DEFAULT_STAT_STATEMENTS_MAX;
+static char	   *stat_statements_exclude_users = NULL;
 
 /*---- Function declarations ----*/
 
@@ -182,6 +230,8 @@ PG_FUNCTION_INFO_V1(statsinfo_cpustats);
 PG_FUNCTION_INFO_V1(statsinfo_cpustats_noarg);
 PG_FUNCTION_INFO_V1(statsinfo_devicestats);
 PG_FUNCTION_INFO_V1(statsinfo_devicestats_noarg);
+PG_FUNCTION_INFO_V1(statsinfo_loadavg);
+PG_FUNCTION_INFO_V1(statsinfo_memory);
 PG_FUNCTION_INFO_V1(statsinfo_profile);
 
 extern Datum PGUT_EXPORT statsinfo_sample(PG_FUNCTION_ARGS);
@@ -194,6 +244,8 @@ extern Datum PGUT_EXPORT statsinfo_cpustats(PG_FUNCTION_ARGS);
 extern Datum PGUT_EXPORT statsinfo_cpustats_noarg(PG_FUNCTION_ARGS);
 extern Datum PGUT_EXPORT statsinfo_devicestats(PG_FUNCTION_ARGS);
 extern Datum PGUT_EXPORT statsinfo_devicestats_noarg(PG_FUNCTION_ARGS);
+extern Datum PGUT_EXPORT statsinfo_loadavg(PG_FUNCTION_ARGS);
+extern Datum PGUT_EXPORT statsinfo_memory(PG_FUNCTION_ARGS);
 extern Datum PGUT_EXPORT statsinfo_profile(PG_FUNCTION_ARGS);
 
 extern PGUT_EXPORT void	_PG_init(void);
@@ -204,23 +256,29 @@ extern PGUT_EXPORT void	fini_last_xact_activity(void);
 /*----  Internal declarations ----*/
 
 static void inet_to_cstring(const SockAddr *addr, char host[NI_MAXHOST]);
+static void StartStatsinfoLauncher(void);
+static void StatsinfoLauncherMain(void);
+static void sil_sigchld_handler(SIGNAL_ARGS);
+static void sil_exit(SIGNAL_ARGS);
 static pid_t exec_background_process(char cmd[]);
 static uint64 get_sysident(void);
 static void must_be_superuser(void);
 static int get_devinfo(const char *path, Datum values[], bool nulls[]);
 static char *get_archive_path(void);
-static const char *elevel_to_str(int elevel);
 static void adjust_log_destination(GucContext context, GucSource source);
 static int get_log_min_messages(void);
 static pid_t get_postmaster_pid(void);
 static bool verify_log_filename(const char *filename);
 static bool verify_timestr(const char *timestr);
+static bool postmaster_is_alive(void);
 
 #if PG_VERSION_NUM >= 90100
 static bool check_textlog_filename(char **newval, void **extra, GucSource source);
+static bool check_enable_maintenance(char **newval, void **extra, GucSource source);
 static bool check_maintenance_time(char **newval, void **extra, GucSource source);
 #else
 static const char *assign_textlog_filename(const char *newval, bool doit, GucSource source);
+static const char *assign_enable_maintenance(const char *newval, bool doit, GucSource source);
 static const char *assign_maintenance_time(const char *newval, bool doit, GucSource source);
 #endif
 
@@ -231,6 +289,12 @@ static int exec_grep(const char *filename, const char *regex, List **records);
 static int exec_split(const char *rawstring, const char *regex, List **fields);
 static bool parse_int64(const char *value, int64 *result);
 static bool parse_float8(const char *value, double *result);
+
+#if PG_VERSION_NUM < 80400
+static bool parse_bool(const char *value, bool *result);
+static const char *elevel_to_str(int elevel);
+#endif
+
 static char *b_trim(char *str);
 static Datum BuildArrayType(List *values, Oid elmtype, Datum(*convert)(void *));
 static Datum _CStringGetTextDatum(void *ptr);
@@ -275,6 +339,10 @@ typedef struct Stats
 
 static Stats	*stats;
 
+/* flags for pg_statsinfo launcher */
+static volatile bool need_exit = false;
+static volatile bool got_SIGCHLD = false;
+
 /*
  * statsinfo_sample - sample statistics for server instance.
  */
@@ -326,6 +394,14 @@ statsinfo_sample(PG_FUNCTION_ARGS)
 			;	/* exclude myself */
 		else if (be->st_waiting)
 			waiting++;
+#if PG_VERSION_NUM >= 90200
+		else if (be->st_state == STATE_IDLE)
+			idle++;
+		else if (be->st_state == STATE_IDLEINTRANSACTION)
+			idle_in_xact++;
+		else if (be->st_state == STATE_RUNNING)
+			running++;
+#else
 		else if (be->st_activity[0] != '\0')
 		{
 			if (strcmp(be->st_activity, "<IDLE>") == 0)
@@ -335,6 +411,7 @@ statsinfo_sample(PG_FUNCTION_ARGS)
 			else
 				running++;
 		}
+#endif
 
 		/*
 		 * sample long transactions, but exclude vacuuming processes.
@@ -349,15 +426,30 @@ statsinfo_sample(PG_FUNCTION_ARGS)
 			continue;
 
 		/* XXX: needs lock? */
-		proc = BackendPidGetProc(be->st_procpid);
-		if (proc == NULL || (proc->vacuumFlags & PROC_IN_VACUUM))
+#if PG_VERSION_NUM >= 90200
+		if ((proc = BackendPidGetProc(be->st_procpid)) == NULL ||
+			(ProcGlobal->allPgXact[proc->pgprocno].vacuumFlags & PROC_IN_VACUUM))
 			continue;
 
+		if (be->st_state == STATE_IDLEINTRANSACTION)
+			strlcpy(stats->max_xact_query,
+				"<IDLE> in transaction", pgstat_track_activity_query_size);
+		else
+			strlcpy(stats->max_xact_query,
+				be->st_activity, pgstat_track_activity_query_size);
+#else
+		if ((proc = BackendPidGetProc(be->st_procpid)) == NULL ||
+			(proc->vacuumFlags & PROC_IN_VACUUM))
+			continue;
+
+		strlcpy(stats->max_xact_query,
+			be->st_activity, pgstat_track_activity_query_size);
+#endif
+
 		stats->max_xact_pid = be->st_procpid;
 		stats->max_xact_start = be->st_xact_start_timestamp;
 		stats->max_xact_duration = duration;
 		inet_to_cstring(&be->st_clientaddr, stats->max_xact_client);
-		strlcpy(stats->max_xact_query, be->st_activity, pgstat_track_activity_query_size);
 	}
 
 	stats->idle += idle;
@@ -621,7 +713,7 @@ _PG_init(void)
 							NULL);
 
 	DefineCustomStringVariable(GUC_PREFIX ".excluded_dbnames",
-							   "Selects which dbnames are excluded by pg_statinfo.",
+							   "Selects which dbnames are excluded by pg_statsinfo.",
 							   NULL,
 							   &excluded_dbnames,
 							   "template0, template1",
@@ -634,7 +726,7 @@ _PG_init(void)
 							   NULL);
 
 	DefineCustomStringVariable(GUC_PREFIX ".excluded_schemas",
-							   "Selects which schema are excluded by pg_statinfo.",
+							   "Selects which schemas are excluded by pg_statsinfo.",
 							   NULL,
 							   &excluded_schemas,
 							   "pg_catalog,pg_toast,information_schema",
@@ -780,18 +872,33 @@ _PG_init(void)
 							   NULL,
 							   NULL);
 
-	DefineCustomBoolVariable(GUC_PREFIX ".enable_maintenance",
-							 "Enable the maintenance.",
-							 NULL,
-							 &enable_maintenance,
-							 DEFAULT_ENABLE_MAINTENANCE,
-							 PGC_SIGHUP,
-							 GUC_SUPERUSER_ONLY,
+	DefineCustomStringVariable(GUC_PREFIX ".textlog_nologging_users",
+							   "Sets dbusers that doesn't logging to textlog.",
+							   NULL,
+							   &textlog_nologging_users,
+							   "",
+							   PGC_SIGHUP,
+							   GUC_SUPERUSER_ONLY,
 #if PG_VERSION_NUM >= 90100
-							 NULL,
+							   NULL,
 #endif
-							 NULL,
-							 NULL);
+							   NULL,
+							   NULL);
+
+	DefineCustomStringVariable(GUC_PREFIX ".enable_maintenance",
+							   "Sets the maintenance mode.",
+							   NULL,
+							   &enable_maintenance,
+							   DEFAULT_ENABLE_MAINTENANCE,
+							   PGC_SIGHUP,
+							   GUC_SUPERUSER_ONLY,
+#if PG_VERSION_NUM >= 90100
+							   check_enable_maintenance,
+							   NULL,
+#else
+						       assign_enable_maintenance,
+#endif
+							   NULL);
 
 	DefineCustomStringVariable(GUC_PREFIX ".maintenance_time",
 							   "Sets the maintenance time.",
@@ -823,6 +930,19 @@ _PG_init(void)
 							NULL,
 							NULL);
 
+	DefineCustomStringVariable(GUC_PREFIX ".log_maintenance_command",
+							   "Sets the shell command that will be called to the log maintenance.",
+							   NULL,
+							   &log_maintenance_command,
+							   DEFAULT_LOG_MAINTENANCE_COMMAND,
+							   PGC_SIGHUP,
+							   0,
+#if PG_VERSION_NUM >= 90100
+							   NULL,
+#endif
+							   NULL,
+							   NULL);
+
 	DefineCustomIntVariable(GUC_PREFIX ".long_lock_threashold",
 							"Sets the threshold of lock wait time.",
 							NULL,
@@ -838,6 +958,34 @@ _PG_init(void)
 							NULL,
 							NULL);
 
+	DefineCustomIntVariable(GUC_PREFIX ".stat_statements_max",
+							"Sets the max collection size from pg_stat_statements.",
+							NULL,
+							&stat_statements_max,
+							DEFAULT_STAT_STATEMENTS_MAX,
+							0,
+							INT_MAX,
+							PGC_SIGHUP,
+							0,
+#if PG_VERSION_NUM >= 90100
+							NULL,
+#endif
+							NULL,
+							NULL);
+
+	DefineCustomStringVariable(GUC_PREFIX ".stat_statements_exclude_users",
+							   "Sets dbusers that doesn't collect statistics of statement from pg_stat_statements.",
+							   NULL,
+							   &stat_statements_exclude_users,
+							   "",
+							   PGC_SIGHUP,
+							   0,
+#if PG_VERSION_NUM >= 90100
+							   NULL,
+#endif
+							   NULL,
+							   NULL);
+
 	EmitWarningsOnPlaceholders("pg_statsinfo");
 
 	if (IsUnderPostmaster)
@@ -893,13 +1041,11 @@ _PG_init(void)
 	init_last_xact_activity();
 #endif
 
-	/* spawn daemon process if the first call */
+	/*
+	 * spawn pg_statsinfo launcher process if the first call
+	 */
 	if (!IsUnderPostmaster)
-	{
-		char	cmd[MAXPGPATH];
-
-		exec_background_process(cmd);
-	}
+		StartStatsinfoLauncher();
 }
 
 /*
@@ -1232,7 +1378,7 @@ get_devicestats(FunctionCallInfo fcinfo, ArrayType *devicestats)
 		/* extract device information */
 		if (exec_grep(FILE_DISKSTATS, regex, &records) <= 0)
 		{
-			ereport(WARNING,
+			ereport(DEBUG2,
 				(errmsg("device information of \"%s\" used by tablespace \"%s\" does not exist in \"%s\"",
 					device, spcname, FILE_DISKSTATS)));
 			prev_device = NULL;
@@ -1407,6 +1553,200 @@ get_devicestats(FunctionCallInfo fcinfo, ArrayType *devicestats)
 	return (Datum) 0;
 }
 
+#define FILE_LOADAVG			"/proc/loadavg"
+#define NUM_LOADAVG_COLS		3
+#define NUM_LOADAVG_FIELDS_MIN	3
+
+/*
+ * statsinfo_loadavg - get loadavg information
+ */
+Datum
+statsinfo_loadavg(PG_FUNCTION_ARGS)
+{
+	TupleDesc	tupdesc;
+	int			fd;
+	char		buffer[256];
+	int			nbytes;
+	float4		loadavg1;
+	float4		loadavg5;
+	float4		loadavg15;
+	HeapTuple	tuple;
+	Datum		values[NUM_LOADAVG_COLS];
+	bool		nulls[NUM_LOADAVG_COLS];
+
+	must_be_superuser();
+
+	/* Build a tuple descriptor for our result type */
+	if (get_call_result_type(fcinfo, NULL, &tupdesc) != TYPEFUNC_COMPOSITE)
+		elog(ERROR, "return type must be a row type");
+
+	Assert(tupdesc->natts == lengthof(values));
+
+	/* extract loadavg information */
+	if ((fd = open(FILE_LOADAVG, O_RDONLY)) < 0)
+		ereport(ERROR,
+			(errcode_for_file_access(),
+			 errmsg("could not open file \"%s\": ", FILE_LOADAVG)));
+
+	if ((nbytes = read(fd, buffer, sizeof(buffer) - 1)) < 0)
+	{
+		close(fd);
+		ereport(ERROR,
+			(errcode_for_file_access(),
+			 errmsg("could not read file \"%s\": ", FILE_LOADAVG)));
+	}
+
+	close(fd);
+	buffer[nbytes] = '\0';
+
+	if (sscanf(buffer, "%f %f %f",
+			&loadavg1, &loadavg5, &loadavg15) < NUM_LOADAVG_FIELDS_MIN)
+		ereport(ERROR,
+			(errcode(ERRCODE_DATA_EXCEPTION),
+			 errmsg("unexpected file format: \"%s\"", FILE_LOADAVG),
+			 errdetail("number of fields is not corresponding")));
+
+	memset(nulls, 0, sizeof(nulls));
+	memset(values, 0, sizeof(values));
+
+	/* loadavg1 */
+	values[0] = Float4GetDatum(loadavg1);
+
+	/* loadavg5 */
+	values[1] = Float4GetDatum(loadavg5);
+
+	/* loadavg15 */
+	values[2] = Float4GetDatum(loadavg15);
+
+	tuple = heap_form_tuple(tupdesc, values, nulls);
+
+	return HeapTupleGetDatum(tuple);
+}
+
+#define FILE_MEMINFO		"/proc/meminfo"
+#define NUM_MEMORY_COLS		5
+
+typedef struct meminfo_table
+{
+	const char	*name;	/* memory type name */
+	int64		*slot;	/* slot in return struct */
+} meminfo_table;
+
+static int
+compare_meminfo_table(const void *a, const void *b)
+{
+	return strcmp(((const meminfo_table *) a)->name, ((const meminfo_table *) b)->name);
+}
+
+/*
+ * statsinfo_memory - get memory information
+ */
+Datum
+statsinfo_memory(PG_FUNCTION_ARGS)
+{
+	TupleDesc		 tupdesc;
+	HeapTuple		 tuple;
+	Datum			 values[NUM_MEMORY_COLS];
+	bool			 nulls[NUM_MEMORY_COLS];
+	int				 fd;
+	char			 buffer[2048];
+	int				 nbytes;
+	int64			 main_free = 0;
+	int64			 buffers = 0;
+	int64			 cached = 0;
+	int64			 swap_free = 0;
+	int64			 swap_total = 0;
+	int64			 dirty = 0;
+	char 			 namebuf[16];
+	char			*head;
+	char			*tail;
+	int				 meminfo_table_count;
+	meminfo_table	 findme = { namebuf, NULL };
+	meminfo_table	*found;
+	meminfo_table	 meminfo_tables[] =
+	{
+		{"Buffers",   &buffers},
+		{"Cached",    &cached},
+		{"Dirty",     &dirty},		/* 2.5.41+ */
+		{"MemFree",   &main_free},
+		{"SwapFree",  &swap_free},
+		{"SwapTotal", &swap_total}
+	};
+
+	must_be_superuser();
+
+	/* Build a tuple descriptor for our result type */
+	if (get_call_result_type(fcinfo, NULL, &tupdesc) != TYPEFUNC_COMPOSITE)
+		elog(ERROR, "return type must be a row type");
+
+	Assert(tupdesc->natts == lengthof(values));
+
+	/* extract memory information */
+	if ((fd = open(FILE_MEMINFO, O_RDONLY)) < 0)
+		ereport(ERROR,
+			(errcode_for_file_access(),
+			 errmsg("could not open file \"%s\": ", FILE_MEMINFO)));
+
+	if ((nbytes = read(fd, buffer, sizeof(buffer) - 1)) < 0)
+	{
+		close(fd);
+		ereport(ERROR,
+			(errcode_for_file_access(),
+			 errmsg("could not read file \"%s\": ", FILE_MEMINFO)));
+	}
+
+	close(fd);
+	buffer[nbytes] = '\0';
+
+	meminfo_table_count = sizeof(meminfo_tables) / sizeof(meminfo_table);
+	head = buffer;
+	for (;;)
+	{
+		if ((tail = strchr(head, ':')) == NULL)
+			break;
+		*tail = '\0';
+		if (strlen(head) >= sizeof(namebuf))
+		{
+			head = tail + 1;
+			goto nextline;
+		}
+		strcpy(namebuf, head);
+		found = bsearch(&findme, meminfo_tables, meminfo_table_count,
+						sizeof(meminfo_table), compare_meminfo_table);
+		head = tail + 1;
+		if (!found)
+			goto nextline;
+		*(found->slot) = strtoul(head, &tail, 10);
+
+nextline:
+		if ((tail = strchr(head, '\n')) == NULL)
+			break;
+		head = tail + 1;
+	}
+
+	memset(nulls, 0, sizeof(nulls));
+	memset(values, 0, sizeof(values));
+
+	/* memfree */
+	values[0] = Int64GetDatum(main_free);
+
+	/* buffers */
+	values[1] = Int64GetDatum(buffers);
+
+	/* cached */
+	values[2] = Int64GetDatum(cached);
+
+	/* swap */
+	values[3] = Int64GetDatum(swap_total - swap_free);
+
+	/* dirty */
+	values[4] = Int64GetDatum(dirty);
+
+	tuple = heap_form_tuple(tupdesc, values, nulls);
+
+	return HeapTupleGetDatum(tuple);
+}
+
 #define FILE_PROFILE		"/proc/systemtap/statsinfo_prof/profile"
 #define NUM_PROFILE_COLS	3
 #define NUM_PROFILE_FIELDS	3
@@ -1580,6 +1920,150 @@ send_u64(int fd, const char *key, uint64 value)
 	send_str(fd, key, buf);
 }
 
+/*
+ * StartStatsinfoLauncher - Main entry point for pg_statsinfo launcher process.
+ */
+static void
+StartStatsinfoLauncher(void)
+{
+	/*
+	 * invoke pg_statsinfo launcher processs
+	 */
+	switch (fork_process())
+	{
+		case -1:
+			ereport(LOG,
+				(errmsg("could not fork pg_statsinfo launcher process: %m")));
+			break;
+		case 0:
+			/* in child process */
+			/* Lose the postmaster's on-exit routines */
+			on_exit_reset();
+
+			StatsinfoLauncherMain();
+			break;
+		default:
+			break;
+	}
+
+	return;
+}
+
+#define LAUNCH_RETRY_PERIOD		300	/* sec */
+#define LAUNCH_RETRY_MAX		5
+
+/*
+ * StatsinfoLauncherMain - Main loop for the pg_statsinfo launcher process.
+ */
+static void
+StatsinfoLauncherMain(void)
+{
+	int			StatsinfoPID;
+	int			launch_retry = 0;
+	pg_time_t	launch_time;
+	char		cmd[MAXPGPATH];
+
+	/* we are postmaster subprocess now */
+	IsUnderPostmaster = true;
+
+	/* Identify myself via ps */
+	init_ps_display("pg_statsinfo launcher process", "", "", "");
+
+	/* delay for the preparation of syslogger */
+	pg_usleep(1000000L);	/* 1s */
+
+	ereport(LOG,
+		(errmsg("pg_statsinfo launcher started")));
+
+	/* Set up signal handlers */
+	pqsignal(SIGHUP, sil_exit);
+	pqsignal(SIGINT, sil_exit);
+	pqsignal(SIGTERM, sil_exit);
+	pqsignal(SIGQUIT, sil_exit);
+	pqsignal(SIGALRM, sil_exit);
+	pqsignal(SIGPIPE, sil_exit);
+	pqsignal(SIGUSR1, sil_exit);
+	pqsignal(SIGUSR2, sil_exit);
+	pqsignal(SIGCHLD, sil_sigchld_handler);
+
+	/* launch a pg_statsinfod process */
+	StatsinfoPID = exec_background_process(cmd);
+	launch_time = (pg_time_t) time(NULL);
+
+	for (;;)
+	{
+		/* pg_statsinfo launcher quits either when the postmaster dies */
+		if (!postmaster_is_alive())
+			break;
+
+		/* have received a signal that terminate process */
+		if (need_exit)
+			break;
+
+		/* pg_statsinfod process died */
+		if (got_SIGCHLD)
+		{
+			int status;
+
+			waitpid(StatsinfoPID, &status, WNOHANG);
+
+			/* pg_statsinfod normally end, terminate the pg_statsinfo launcher */
+			if (status == 0)
+				break;
+
+			/* 
+			 * if the pg_statsinfod was aborted with fatal error,
+			 * then terminate the pg_statsinfo launcher
+			 */
+			if (WIFEXITED(status) && WEXITSTATUS(status) == STATSINFO_EXIT_FAILED)
+			{
+				ereport(WARNING,
+					(errmsg("pg_statsinfod is aborted with fatal error, "
+							"terminate the pg_statsinfo launcher")));
+				break;
+			}
+
+			/* pg_statsinfod abnormally end, relaunch new pg_statsinfod process */
+			ereport(WARNING,
+				(errmsg("pg_statsinfod is aborted")));
+
+			/* 
+			 * if the pg_statsinfod was aborted continuously,
+			 * then terminate the pg_statsinfo launcher
+			 */
+			if (((pg_time_t) time(NULL) - launch_time) <= LAUNCH_RETRY_PERIOD)
+			{
+				if (launch_retry >= LAUNCH_RETRY_MAX)
+				{
+					ereport(WARNING,
+					(errmsg("pg_statsinfod is aborted continuously, "
+							"terminate the pg_statsinfo launcher")));
+					break;
+				}
+			}
+			else
+				launch_retry = 0;
+
+			ereport(LOG,
+				(errmsg("relaunch a pg_statsinfod process")));
+
+			got_SIGCHLD = false;
+			StatsinfoPID = exec_background_process(cmd);
+			launch_time = (pg_time_t) time(NULL);
+
+			launch_retry++;
+		}
+
+		pg_usleep(100000L);		/* 100ms */
+	}
+
+	/* Normal exit from the pg_statsinfo launcher is here */
+	ereport(LOG,
+		(errmsg("pg_statsinfo launcher shutting down")));
+
+	proc_exit(0);
+}
+
 /*
  * exec_background_process - Start statsinfo background process.
  */
@@ -1587,18 +2071,15 @@ static pid_t
 exec_background_process(char cmd[])
 {
 	char		binpath[MAXPGPATH];
-	char		logpath[MAXPGPATH];
 	char		share_path[MAXPGPATH];
-	char		prev_csv_name[MAXPGPATH];
 	uint64		sysident;
 	int			fd;
 	pid_t		fpid;
 	pid_t		postmaster_pid = get_postmaster_pid();
-	pg_time_t	now;
+	pg_time_t	log_ts;
 	pg_tz	   *log_tz;
-	int			len;
 
-	now = (pg_time_t) time(NULL) - 1;
+	log_ts = (pg_time_t) time(NULL);
 	log_tz = pg_tzset(GetConfigOption("log_timezone", false));
 
 	/* $PGHOME/bin */
@@ -1608,20 +2089,6 @@ exec_background_process(char cmd[])
 	/* $PGHOME/share */
 	get_share_path(my_exec_path, share_path);
 
-	/* $PGDATA/pg_log */
-	if (is_absolute_path(Log_directory))
-		strlcpy(logpath, Log_directory, MAXPGPATH);
-	else
-		join_path_components(logpath, DataDir, Log_directory);
-
-	/* prev_csv_name */
-	pg_strftime(prev_csv_name, MAXPGPATH, Log_filename,
-				pg_localtime(&now, log_tz));
-	len = strlen(prev_csv_name);
-	if (len > 4 && (strcmp(prev_csv_name + (len - 4), ".log") == 0))
-		len -= 4;
-	strlcpy(prev_csv_name + len, ".csv", MAXPGPATH - len);
-
 	/* ControlFile: system_identifier */
 	sysident = get_sysident();
 
@@ -1639,39 +2106,10 @@ exec_background_process(char cmd[])
 	send_str(fd, "port", GetConfigOption("port", false));
 	send_str(fd, "server_version_num", GetConfigOption("server_version_num", false));
 	send_str(fd, "server_version_string", GetConfigOption("server_version", false));
-	send_str(fd, "prev_csv_name", prev_csv_name);
 	send_str(fd, "share_path", share_path);
 	send_i32(fd, "server_encoding", GetDatabaseEncoding());
 	send_str(fd, "data_directory", DataDir);
-	send_str(fd, "log_timezone", pg_localtime(&now, log_tz)->tm_zone);
-	send_str(fd, "log_directory", Log_directory);
-	send_str(fd, "log_error_verbosity", GetConfigOption("log_error_verbosity", false));
-	send_str(fd, GUC_PREFIX ".syslog_min_messages", elevel_to_str(syslog_min_messages));
-	send_str(fd, GUC_PREFIX ".textlog_min_messages", elevel_to_str(textlog_min_messages));
-	send_str(fd, GUC_PREFIX ".textlog_filename", textlog_filename);
-	send_str(fd, GUC_PREFIX ".textlog_line_prefix", textlog_line_prefix);
-	send_str(fd, GUC_PREFIX ".syslog_line_prefix", syslog_line_prefix);
-	send_i32(fd, GUC_PREFIX ".textlog_permission", textlog_permission);
-	send_str(fd, GUC_PREFIX ".excluded_dbnames", excluded_dbnames);
-	send_str(fd, GUC_PREFIX ".excluded_schemas", excluded_schemas);
-	send_i32(fd, GUC_PREFIX ".sampling_interval", sampling_interval);
-	send_i32(fd, GUC_PREFIX ".snapshot_interval", snapshot_interval);
-	send_str(fd, GUC_PREFIX ".repository_server", repository_server);
-	send_i32(fd, GUC_PREFIX ".adjust_log_level", adjust_log_level);
-	send_str(fd, GUC_PREFIX ".adjust_log_info", adjust_log_info);
-	send_str(fd, GUC_PREFIX ".adjust_log_notice", adjust_log_notice);
-	send_str(fd, GUC_PREFIX ".adjust_log_warning", adjust_log_warning);
-	send_str(fd, GUC_PREFIX ".adjust_log_error", adjust_log_error);
-	send_str(fd, GUC_PREFIX ".adjust_log_log", adjust_log_log);
-	send_str(fd, GUC_PREFIX ".adjust_log_fatal", adjust_log_fatal);
-	send_i32(fd, GUC_PREFIX ".enable_maintenance", enable_maintenance);
-	send_str(fd, GUC_PREFIX ".maintenance_time", maintenance_time);
-	send_i32(fd, GUC_PREFIX ".repository_keepday", repository_keepday);
-
-#ifdef HAVE_SYSLOG
-	send_str(fd, "syslog_facility", GetConfigOption("syslog_facility", false));
-	send_str(fd, "syslog_ident", GetConfigOption("syslog_ident", false));
-#endif
+	send_str(fd, "log_timezone", pg_localtime(&log_ts, log_tz)->tm_zone);
 	send_str(fd, ":debug", _("DEBUG"));
 	send_str(fd, ":info", _("INFO"));
 	send_str(fd, ":notice", _("NOTICE"));
@@ -1695,6 +2133,19 @@ exec_background_process(char cmd[])
 	return fpid;
 }
 
+/* SIGCHLD: pg_statsinfod process died */
+static void
+sil_sigchld_handler(SIGNAL_ARGS)
+{
+	got_SIGCHLD = true;
+}
+
+static void
+sil_exit(SIGNAL_ARGS)
+{
+	need_exit = true;
+}
+
 /*
  * Read control file. We cannot retrieve it from "Control File" shared memory
  * because the shared memory might not be initialized yet.
@@ -1780,7 +2231,6 @@ statsinfo_tablespaces(PG_FUNCTION_ARGS)
 	{
 		Form_pg_tablespace form = (Form_pg_tablespace) GETSTRUCT(tuple);
 		Datum			datum;
-		bool			isnull;
 
 		memset(values, 0, sizeof(values));
 		memset(nulls, 0, sizeof(nulls));
@@ -1793,28 +2243,31 @@ statsinfo_tablespaces(PG_FUNCTION_ARGS)
 		values[i++] = CStringGetTextDatum(NameStr(form->spcname));
 
 		/* location */
-		datum = fastgetattr(tuple, Anum_pg_tablespace_spclocation,
-							RelationGetDescr(relation), &isnull);
-		if (isnull || VARSIZE_ANY_EXHDR(datum) == 0)
-		{
-			path = DataDir;
+		if (HeapTupleGetOid(tuple) == DEFAULTTABLESPACE_OID ||
+			HeapTupleGetOid(tuple) == GLOBALTABLESPACE_OID)
 			datum = CStringGetTextDatum(DataDir);
-		}
 		else
 		{
-			path = TextDatumGetCString(datum);
+#if PG_VERSION_NUM >= 90200
+			datum = DirectFunctionCall1(pg_tablespace_location,
+										ObjectIdGetDatum(HeapTupleGetOid(tuple)));
+#else
+			bool isnull;
+			datum = fastgetattr(tuple, Anum_pg_tablespace_spclocation,
+								RelationGetDescr(relation), &isnull);
+			/* resolve symlink */
+			if ((len = readlink(TextDatumGetCString(datum),
+								location, lengthof(location))) > 0)
+			{
+				location[len] = '\0';
+				datum = CStringGetTextDatum(location);
+			}
+#endif
 		}
 		values[i++] = datum;
 
-		/* resolve symlink */
-		if ((len = readlink(path, location, lengthof(location))) > 0)
-		{
-			location[len] = '\0';
-			path = location;
-		}
-
 		/* device */
-		i += get_devinfo(path, values + i, nulls + i);
+		i += get_devinfo(TextDatumGetCString(datum), values + i, nulls + i);
 
 		/* spcoptions */
 #if PG_VERSION_NUM >= 90000
@@ -1954,22 +2407,6 @@ get_archive_path(void)
 	return NULL;
 }
 
-static const char *
-elevel_to_str(int elevel)
-{
-	const struct config_enum_entry *e;
-
-	for (e = elevel_options; e && e->name; e++)
-	{
-		if (e->val == elevel)
-			return e->name;
-	}
-
-	elog(ERROR, "could not find enum option %d for %s",
-		 elevel, GUC_PREFIX ".log_min_messages");
-	return NULL;				/* silence compiler */
-}
-
 /*
  * Remove 'stderr' and add 'csvlog' to log_destination.
  */
@@ -2099,6 +2536,66 @@ check_textlog_filename(char **newval, void **extra, GucSource source)
 	return true;
 }
 
+/* forbid unrecognized keyword for maintenance mode */
+static bool
+check_enable_maintenance(char **newval, void **extra, GucSource source)
+{
+	char		*rawstring;
+	List		*elemlist;
+	ListCell	*cell;
+	bool		 bool_val;
+	int			 mode = 0x00;
+	char		 mode_string[32];
+
+	if (parse_bool(*newval, &bool_val))
+	{
+		if (bool_val)
+		{
+			mode |= MAINTENANCE_MODE_SNAPSHOT;
+			mode |= MAINTENANCE_MODE_LOG;
+		}
+		snprintf(mode_string, sizeof(mode_string), "%d", mode);
+		*newval = strdup(mode_string);
+		return true;
+	}
+
+	/* Need a modifiable copy of string */
+	rawstring = pstrdup(*newval);
+
+	if (!SplitIdentifierString(rawstring, ',', &elemlist))
+	{
+		GUC_check_errdetail(GUC_PREFIX ".enable_maintenance list syntax is invalid");
+		goto error;
+	}
+
+	foreach(cell, elemlist)
+	{
+		char *tok = (char *) lfirst(cell);
+
+		if (pg_strcasecmp(tok, "snapshot") == 0)
+			mode |= MAINTENANCE_MODE_SNAPSHOT;
+		else if (pg_strcasecmp(tok, "log") == 0)
+			mode |= MAINTENANCE_MODE_LOG;
+		else
+		{
+			GUC_check_errdetail(GUC_PREFIX ".enable_maintenance unrecognized keyword: \"%s\"", tok);
+			goto error;
+		}
+	}
+
+	pfree(rawstring);
+	list_free(elemlist);
+
+	snprintf(mode_string, sizeof(mode_string), "%d", mode);
+	*newval = strdup(mode_string);
+	return true;
+
+error:
+	pfree(rawstring);
+	list_free(elemlist);
+	return false;
+}
+
 /* forbid empty and invalid time format */
 static bool
 check_maintenance_time(char **newval, void **extra, GucSource source)
@@ -2142,6 +2639,67 @@ assign_textlog_filename(const char *newval, bool doit, GucSource source)
 	return newval;
 }
 
+/* forbid unrecognized keyword for maintenance mode */
+static const char *
+assign_enable_maintenance(const char *newval, bool doit, GucSource source)
+{
+	char		*rawstring;
+	List		*elemlist;
+	ListCell	*cell;
+	bool		 bool_val;
+	int			 mode = 0x00;
+	char		 mode_string[32];
+
+	if (parse_bool(newval, &bool_val))
+	{
+		if (bool_val)
+		{
+			mode |= MAINTENANCE_MODE_SNAPSHOT;
+			mode |= MAINTENANCE_MODE_LOG;
+		}
+		snprintf(mode_string, sizeof(mode_string), "%d", mode);
+		return strdup(mode_string);
+	}
+
+	/* Need a modifiable copy of string */
+	rawstring = pstrdup(newval);
+
+	if (!SplitIdentifierString(rawstring, ',', &elemlist))
+	{
+		pfree(rawstring);
+		list_free(elemlist);
+		ereport(GUC_complaint_elevel(source),
+				(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+				 errmsg(GUC_PREFIX ".enable_maintenance list syntax is invalid")));
+		return NULL;
+	}
+
+	foreach(cell, elemlist)
+	{
+		char *tok = (char *) lfirst(cell);
+
+		if (pg_strcasecmp(tok, "snapshot") == 0)
+			mode |= MAINTENANCE_MODE_SNAPSHOT;
+		else if (pg_strcasecmp(tok, "log") == 0)
+			mode |= MAINTENANCE_MODE_LOG;
+		else
+		{
+			pfree(rawstring);
+			list_free(elemlist);
+			ereport(GUC_complaint_elevel(source),
+				(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+				 errmsg(GUC_PREFIX ".enable_maintenance unrecognized keyword: \"%s\"", tok)));
+			return NULL;
+		}
+	}
+
+	pfree(rawstring);
+	list_free(elemlist);
+
+	snprintf(mode_string, sizeof(mode_string), "%d", mode);
+	return strdup(mode_string);
+}
+
 /* forbid empty and invalid time format */
 static const char *
 assign_maintenance_time(const char *newval, bool doit, GucSource source)
@@ -2218,6 +2776,22 @@ str_to_elevel(const char *name,
 			 errmsg("invalid value for parameter \"%s\": \"%s\"", name, str)));
 	return 0;
 }
+
+static const char *
+elevel_to_str(int elevel)
+{
+	const struct config_enum_entry *e;
+
+	for (e = elevel_options; e && e->name; e++)
+	{
+		if (e->val == elevel)
+			return e->name;
+	}
+
+	elog(ERROR, "could not find enum option %d for %s",
+		 elevel, GUC_PREFIX ".log_min_messages");
+	return NULL;				/* silence compiler */
+}
 #endif
 
 #if PG_VERSION_NUM < 80400
@@ -2498,6 +3072,73 @@ parse_float8(const char *value, double *result)
 	return true;
 }
 
+#if PG_VERSION_NUM < 80400
+/*
+ * Try to interpret value as boolean value.  Valid values are: true,
+ * false, yes, no, on, off, 1, 0; as well as unique prefixes thereof.
+ * If the string parses okay, return true, else false.
+ * If okay and result is not NULL, return the value in *result.
+ */
+static bool
+parse_bool(const char *value, bool *result)
+{
+	size_t		len = strlen(value);
+
+	if (pg_strncasecmp(value, "true", len) == 0)
+	{
+		if (result)
+			*result = true;
+	}
+	else if (pg_strncasecmp(value, "false", len) == 0)
+	{
+		if (result)
+			*result = false;
+	}
+
+	else if (pg_strncasecmp(value, "yes", len) == 0)
+	{
+		if (result)
+			*result = true;
+	}
+	else if (pg_strncasecmp(value, "no", len) == 0)
+	{
+		if (result)
+			*result = false;
+	}
+
+	/* 'o' is not unique enough */
+	else if (pg_strncasecmp(value, "on", (len > 2 ? len : 2)) == 0)
+	{
+		if (result)
+			*result = true;
+	}
+	else if (pg_strncasecmp(value, "off", (len > 2 ? len : 2)) == 0)
+	{
+		if (result)
+			*result = false;
+	}
+
+	else if (pg_strcasecmp(value, "1") == 0)
+	{
+		if (result)
+			*result = true;
+	}
+	else if (pg_strcasecmp(value, "0") == 0)
+	{
+		if (result)
+			*result = false;
+	}
+
+	else
+	{
+		if (result)
+			*result = false; /* suppress compiler warning */
+		return false;
+	}
+	return true;
+}
+#endif
+
 /*
  * Note: this function modify the argument string
  */
@@ -2587,3 +3228,38 @@ search_devicestats(ArrayType *devicestats, const char *device_name)
 	/* not found */
 	return NULL;
 }
+
+/*
+ * postmaster_is_alive - check whether postmaster process is still alive
+ */
+static bool
+postmaster_is_alive(void)
+{
+#ifndef WIN32
+	pid_t	ppid = getppid();
+
+	/* If the postmaster is still our parent, it must be alive. */
+	if (ppid == PostmasterPid)
+		return true;
+
+	/* If the init process is our parent, postmaster must be dead. */
+	if (ppid == 1)
+		return false;
+
+	/*
+	 * If we get here, our parent process is neither the postmaster nor init.
+	 * This can occur on BSD and MacOS systems if a debugger has been attached.
+	 * We fall through to the less-reliable kill() method.
+	 */
+
+	/*
+	 * Use kill() to see if the postmaster is still alive. This can sometimes
+	 * give a false positive result, since the postmaster's PID may get
+	 * recycled, but it is good enough for existing uses by indirect children
+	 * and in debugging environments.
+	 */
+	return (kill(PostmasterPid, 0) == 0);
+#else							/* WIN32 */
+	return (WaitForSingleObject(PostmasterHandle, 0) == WAIT_TIMEOUT);
+#endif   /* WIN32 */
+}
