@@ -1,12 +1,12 @@
 /*
  * pg_statsinfo.c
  *
- * Copyright (c) 2011, NIPPON TELEGRAPH AND TELEPHONE CORPORATION
+ * Copyright (c) 2011-2012, NIPPON TELEGRAPH AND TELEPHONE CORPORATION
  */
 
 #include "pg_statsinfo.h"
 
-const char *PROGRAM_VERSION	= "2.3.1";
+const char *PROGRAM_VERSION	= "2.4.0";
 const char *PROGRAM_URL		= "http://pgstatsinfo.projects.postgresql.org/";
 const char *PROGRAM_EMAIL	= "pgstatsinfo-general@pgfoundry.org";
 
@@ -128,7 +128,7 @@ pgut_help(bool details)
 	printf("                          * CheckpointActivity\n");
 	printf("                          * AutovacuumActivity\n");
 	printf("                          * QueryActivity\n");
-	printf("                          * LockActivity\n");
+	printf("                          * LockConflicts\n");
 	printf("                          * ReplicationActivity\n");
 	printf("                          * SettingParameters\n");
 	printf("                          * SchemaInformation\n");
